from enum import Enum
from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.commands import DownloadCommand
from cognite_toolkit._cdf_tk.constants import DATA_DEFAULT_DIR
from cognite_toolkit._cdf_tk.storageio import (
    AssetIO,
    HierarchyIO,
    InstanceIO,
    RawIO,
)
from cognite_toolkit._cdf_tk.storageio.selectors import (
    AssetCentricSelector,
    AssetSubtreeSelector,
    DataSetSelector,
    InstanceSpaceSelector,
    RawTableSelector,
    SelectedTable,
    SelectedView,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetInteractiveSelect,
    DataModelingSelect,
    RawTableInteractiveSelect,
)


class RawFormats(str, Enum):
    ndjson = "ndjson"
    yaml = "yaml"


class AssetCentricFormats(str, Enum):
    csv = "csv"
    parquet = "parquet"
    ndjson = "ndjson"


class HierarchyFormats(str, Enum):
    ndjson = "ndjson"


class InstanceFormats(str, Enum):
    ndjson = "ndjson"


class InstanceTypes(str, Enum):
    node = "node"
    edge = "edge"


class CompressionFormat(str, Enum):
    gzip = "gzip"
    none = "none"


DEFAULT_DOWNLOAD_DIR = Path(DATA_DEFAULT_DIR)


class DownloadApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.download_main)
        self.command("raw")(self.download_raw_cmd)
        self.command("assets")(self.download_assets_cmd)
        self.command("hierarchy")(self.download_hierarchy_cmd)
        self.command("instances")(self.download_instances_cmd)

    @staticmethod
    def download_main(ctx: typer.Context) -> None:
        """Commands to download data from CDF into a temporary directory."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf download --help[/] for more information.")
        return None

    @staticmethod
    def download_raw_cmd(
        ctx: typer.Context,
        tables: Annotated[
            list[str] | None,
            typer.Argument(
                help="List of tables to download. If not provided, an interactive selection will be made.",
            ),
        ] = None,
        database: Annotated[
            str | None,
            typer.Option(
                "--database",
                "-d",
                help="Database to download from. If not provided, the user will be prompted to select a database.",
            ),
        ] = None,
        file_format: Annotated[
            RawFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the raw tables in. Supported formats: ndjson, yaml",
            ),
        ] = RawFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the raw tables. Supported formats: gzip, none.",
            ),
        ] = CompressionFormat.gzip,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the raw tables.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum the number of records to download from each table. Use -1 to download all records.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download RAW tables from CDF into a temporary directory."""
        cmd = DownloadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        if tables and database:
            selectors = [RawTable(db_name=database, table_name=table) for table in tables]
        elif tables and not database:
            raise typer.BadParameter(
                "The '--database' option is required when specifying tables as arguments.",
                param_hint="--database",
            )
        elif not tables and database:
            selectors = RawTableInteractiveSelect(client, "download").select_tables(database=database)
        else:
            selectors = RawTableInteractiveSelect(client, "download").select_tables()

        cmd.run(
            lambda: cmd.download(
                selectors=[
                    RawTableSelector(table=SelectedTable(db_name=item.db_name, table_name=item.table_name))
                    for item in selectors
                ],
                io=RawIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_assets_cmd(
        ctx: typer.Context,
        data_sets: Annotated[
            list[str] | None,
            typer.Option(
                "--data-set",
                "-d",
                help="List of data sets to download assets from. If this and hierarchy are not provided, an interactive selection will be made.",
            ),
        ] = None,
        hierarchy: Annotated[
            list[str] | None,
            typer.Option(
                "--hierarchy",
                "-r",
                help="List of asset hierarchies to download assets from. If this and data sets are not provided, an interactive selection will be made.",
            ),
        ] = None,
        file_format: Annotated[
            AssetCentricFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the assets in.",
            ),
        ] = AssetCentricFormats.csv,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the assets.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the assets.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of assets to download from each dataset/hierarchy. Use -1 to download all assets.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download assets from CDF into a temporary directory."""
        client = EnvironmentVariables.create_from_environment().get_client()
        is_interactive = not data_sets and not hierarchy
        if is_interactive:
            interactive = AssetInteractiveSelect(client, "download assets")
            selector_type = interactive.select_hierarchies_or_data_sets()
            if selector_type == "Data Set":
                data_sets = interactive.select_data_sets()
            else:
                hierarchy = interactive.select_hierarchies()

        selectors: list[AssetCentricSelector] = []
        if data_sets:
            selectors.extend([DataSetSelector(data_set_external_id=ds, kind="Assets") for ds in data_sets])
        if hierarchy:
            selectors.extend([AssetSubtreeSelector(hierarchy=h, kind="Assets") for h in hierarchy])
        cmd = DownloadCommand()
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=AssetIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_hierarchy_cmd(
        ctx: typer.Context,
        hierarchy: Annotated[
            str | None,
            typer.Argument(
                help="The asset hierarchy to download.",
            ),
        ] = None,
        file_format: Annotated[
            HierarchyFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format for downloading the asset hierarchy.",
            ),
        ] = HierarchyFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the assets.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the asset hierarchy.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum number of resources to download for each type. Use -1 to download all assets.",
            ),
        ] = 100_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download an asset hierarchy from CDF into a temporary directory."""
        cmd = DownloadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        if hierarchy is None:
            selector = AssetInteractiveSelect(client, "download")
            hierarchy = selector.select_hierarchy(allow_empty=False)

        selectors = [
            # MyPy cannot see that resource_type is one of the allowed literals.
            AssetSubtreeSelector(hierarchy=hierarchy, kind=resource_type)  # type: ignore[arg-type]
            for resource_type in ["Assets", "Events", "FileMetadata", "TimeSeries"]
        ]
        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=HierarchyIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )

    @staticmethod
    def download_instances_cmd(
        ctx: typer.Context,
        instance_space: Annotated[
            str | None,
            typer.Option(
                "--instance-space",
                "-s",
                help="The instance space to download instances from. If not provided, an interactive "
                "selection will be made.",
            ),
        ] = None,
        schema_space: Annotated[
            str | None,
            typer.Option(
                "--schema-space",
                "-c",
                help="The schema space where the views are located.",
            ),
        ] = None,
        view_external_ids: Annotated[
            list[str] | None,
            typer.Option(
                "--view",
                "-w",
                help="List of view external IDs to download properties for the "
                "instances. To specify version use a forward slash, e.g. viewExternalId/v1.",
            ),
        ] = None,
        instance_type: Annotated[
            InstanceTypes,
            typer.Option(
                "--instance-type",
                "-t",
                help="The type of instances to download.",
            ),
        ] = InstanceTypes.node,
        file_format: Annotated[
            InstanceFormats,
            typer.Option(
                "--format",
                "-f",
                help="Format to download the instances in.",
            ),
        ] = InstanceFormats.ndjson,
        compression: Annotated[
            CompressionFormat,
            typer.Option(
                "--compression",
                "-z",
                help="Compression format to use when downloading the instances.",
            ),
        ] = CompressionFormat.none,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Where to download the instances.",
                allow_dash=True,
            ),
        ] = DEFAULT_DOWNLOAD_DIR,
        limit: Annotated[
            int,
            typer.Option(
                "--limit",
                "-l",
                help="The maximum the number of instances to download from each view. Use -1 to download all.",
            ),
        ] = 10_000,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will download Instances from CDF into a temporary directory."""
        cmd = DownloadCommand()

        client = EnvironmentVariables.create_from_environment().get_client()
        if instance_space is None:
            selector = DataModelingSelect(client, "download instances")
            selected_instance_space = selector.select_instance_space(multiselect=False)
            selected_instance_type = selector.select_instance_type()
            selected_schema_space = selector.select_schema_space(
                include_global=True, message="In which space is the views with instance properties located?"
            ).space
            selected_views = selector.select_view(
                multiselect=True,
                space=selected_schema_space,
                message="Select views to download instance properties from.",
                include_global=True,
                instance_type=selected_instance_type,
            )
            selectors: list[InstanceSpaceSelector] = [
                InstanceSpaceSelector(
                    instance_space=selected_instance_space,
                    view=SelectedView(
                        space=selected_schema_space,
                        external_id=view.external_id,
                        version=view.version,
                    ),
                    instance_type=selected_instance_type,
                )
                for view in selected_views
            ]
        elif schema_space is None and view_external_ids is None:
            selectors = [InstanceSpaceSelector(instance_space=instance_space, instance_type=instance_type.value)]
        elif schema_space is not None and view_external_ids is not None:
            selectors = [
                InstanceSpaceSelector(
                    instance_space=instance_space,
                    view=SelectedView(
                        space=schema_space,
                        external_id=view_id_str.split("/", maxsplit=1)[0],
                        version=view_id_str.split("/", maxsplit=1)[1] if "/" in view_id_str else None,
                    ),
                    instance_type=instance_type.value,
                )
                for view_id_str in view_external_ids
            ]
        else:
            raise typer.BadParameter(
                "Both '--schema-space' and '--view' must be provided together.",
                param_hint="--view",
            )

        cmd.run(
            lambda: cmd.download(
                selectors=selectors,
                io=InstanceIO(client),
                output_dir=output_dir,
                file_format=f".{file_format.value}",
                compression=compression.value,
                limit=limit if limit != -1 else None,
                verbose=verbose,
            )
        )



import re
from typing import Literal

from pydantic import Field, field_validator, model_serializer
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS,
    FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER,
    SPACE_FORMAT_PATTERN,
)
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import ToolkitResource
from .container_field_definitions import ConstraintDefinition, ContainerPropertyDefinition, IndexDefinition

KEY_PATTERN = re.compile(CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN)


class ContainerYAML(ToolkitResource):
    space: str = Field(
        description="The workspace for the container, a unique identifier for the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="name for the container.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the container.",
        max_length=1024,
    )
    used_for: Literal["node", "edge", "all"] | None = Field(
        default=None,
        description="Should this operation apply to nodes, edges or both.",
    )
    properties: dict[str, ContainerPropertyDefinition] = Field(
        description="Set of properties to apply to the container."
    )
    constraints: dict[str, ConstraintDefinition] | None = Field(
        default=None,
        description="Set of constraints to apply to the container.",
    )
    indexes: dict[str, IndexDefinition] | None = Field(
        default=None,
        description="Set of indexes to apply to the container.",
        max_length=10,
    )

    @field_validator("external_id")
    @classmethod
    def check_forbidden_external_id_value(cls, val: str) -> str:
        """Check the external_id not present in forbidden set"""
        if val in FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS:
            raise ValueError(
                f"'{val}' is a reserved container External ID. Reserved External IDs are: {humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS)}"
            )
        return val

    @field_validator("properties")
    @classmethod
    def validate_properties_identifier(cls, val: dict[str, str]) -> dict[str, str]:
        """Validate properties Identifier"""
        for key in val.keys():
            if not KEY_PATTERN.match(key):
                raise ValueError(f"Property '{key}' does not match the required pattern: {KEY_PATTERN.pattern}")
            if key in FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER:
                raise ValueError(
                    f"'{key}' is a reserved property identifier. Reserved identifiers are: {humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER)}"
                )
        return val

    @model_serializer(mode="wrap")
    def serialize_container(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        serialized_data = handler(self)
        if self.constraints:
            serialized_data["constraints"] = {k: v.model_dump(**vars(info)) for k, v in self.constraints.items()}
        if self.indexes:
            serialized_data["indexes"] = {k: v.model_dump(**vars(info)) for k, v in self.indexes.items()}
        return serialized_data



import errno
import os
import re
import shutil
import stat
import tempfile
import time
import typing
import warnings
from abc import abstractmethod
from collections import UserDict, defaultdict
from collections.abc import Hashable, ItemsView, KeysView, ValuesView
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Literal, TypeVar, overload
from zipfile import ZipFile

import pandas as pd
import yaml
from rich import print

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import ENV_VAR_PATTERN, HINT_LEAD_TEXT, URL
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitValueError,
    ToolkitYAMLFormatError,
)
from cognite_toolkit._cdf_tk.tk_warnings import EnvironmentVariableMissingWarning, MediumSeverityWarning


@overload
def load_yaml_inject_variables(
    filepath: Path | str,
    environment_variables: dict[str, str | None],
    required_return_type: Literal["list"],
    validate: bool = True,
    original_filepath: Path | None = None,
) -> list[dict[str, Any]]: ...


@overload
def load_yaml_inject_variables(
    filepath: Path | str,
    environment_variables: dict[str, str | None],
    required_return_type: Literal["dict"],
    validate: bool = True,
    original_filepath: Path | None = None,
) -> dict[str, Any]: ...


@overload
def load_yaml_inject_variables(
    filepath: Path | str,
    environment_variables: dict[str, str | None],
    required_return_type: Literal["any", "list", "dict"] = "any",
    validate: bool = True,
    original_filepath: Path | None = None,
) -> dict[str, Any] | list[dict[str, Any]]: ...


def load_yaml_inject_variables(
    filepath: Path | str,
    environment_variables: dict[str, str | None],
    required_return_type: Literal["any", "list", "dict"] = "any",
    validate: bool = True,
    original_filepath: Path | None = None,
) -> dict[str, Any] | list[dict[str, Any]]:
    """Loads a YAML file and injects environment variables into it.

    Args:
        filepath (Path | str): Path to the YAML file or file content.
        environment_variables (dict[str, str | None]): Dictionary with environment variables.
        required_return_type (Literal["any", "list", "dict"], optional): The required return type. Defaults to "any".
        validate (bool, optional): Whether to validate that all environment variables were replaced. Defaults to True.
        original_filepath (Path | None, optional): In case the filepath is a string, this is the original path.
            Used for error messages. Defaults to None.

    Returns:
        dict[str, Any] | list[dict[str, Any]]: The YAML content with the environment variables injected.

    """
    if isinstance(filepath, str):
        content = filepath
    else:
        content = safe_read(filepath)
    for key, value in environment_variables.items():
        if value is None:
            continue
        content = content.replace(f"${{{key}}}", str(value))
    if validate and (missing_variables := [match.group(1) for match in ENV_VAR_PATTERN.finditer(content)]):
        if isinstance(filepath, Path):
            source = filepath
        elif original_filepath:
            source = original_filepath
        else:
            source = Path("UNKNOWN")
        warnings.warn(EnvironmentVariableMissingWarning(source, frozenset(missing_variables)), stacklevel=2)

    result = read_yaml_content(content)

    if required_return_type == "any":
        return result
    elif required_return_type == "list":
        if isinstance(result, list):
            return result
        raise ValueError(f"Expected a list, but got {type(result)}")
    elif required_return_type == "dict":
        if isinstance(result, dict):
            return result
        raise ValueError(f"Expected a dict, but got {type(result)}")
    else:
        raise ValueError(f"Unknown required_return_type {required_return_type}")


@overload
def read_yaml_file(filepath: Path, expected_output: Literal["dict"] = "dict") -> dict[str, Any]: ...


@overload
def read_yaml_file(filepath: Path, expected_output: Literal["list"]) -> list[dict[str, Any]]: ...


def read_yaml_file(
    filepath: Path, expected_output: Literal["list", "dict"] = "dict"
) -> dict[str, Any] | list[dict[str, Any]]:
    """Read a YAML file and return a dictionary

    filepath: path to the YAML file
    """
    try:
        config_data = read_yaml_content(safe_read(filepath))
    except yaml.YAMLError as e:
        print(f"  [bold red]ERROR:[/] reading {filepath}: {e}")
        return {}

    if expected_output == "list" and isinstance(config_data, dict):
        ToolkitYAMLFormatError(f"{filepath} did not contain `list` as expected")
    elif expected_output == "dict" and isinstance(config_data, list):
        ToolkitYAMLFormatError(f"{filepath} did not contain `dict` as expected")
    return config_data


_TOTAL_ELAPSED_TIME = 0.0
_HAS_HINTED = False


def keyvault_constructor(loader: yaml.Loader, node: yaml.Node) -> Any:
    """Constructor to preserve the !keyvault tag along with the value."""
    if isinstance(node, yaml.ScalarNode):
        value = loader.construct_scalar(node)
        return f"!keyvault {value}"
    elif isinstance(node, yaml.MappingNode):
        return loader.construct_mapping(node)
    elif isinstance(node, yaml.SequenceNode):
        return loader.construct_sequence(node)
    else:
        # This should never happen as YAML only has these three node types
        # But we need this for type checking completeness
        raise ValueError(f"Unexpected node type: {type(node)}")


yaml.SafeLoader.add_constructor("!keyvault", keyvault_constructor)

# Register for the fast CSafeLoader (used directly when libyaml is available)
if hasattr(yaml, "CSafeLoader"):
    yaml.CSafeLoader.add_constructor("!keyvault", keyvault_constructor)


def read_yaml_content(content: str) -> dict[str, Any] | list[dict[str, Any]]:
    """Read a YAML string and return a dictionary

    content: string containing the YAML content
    """
    global _TOTAL_ELAPSED_TIME, _HAS_HINTED
    if yaml.__with_libyaml__:
        # CSafeLoader is faster than yaml.safe_load
        return yaml.CSafeLoader(content).get_data()

    t0 = time.perf_counter()
    result = yaml.safe_load(content)
    _TOTAL_ELAPSED_TIME += time.perf_counter() - t0
    if _TOTAL_ELAPSED_TIME > 60.0 and not _HAS_HINTED:
        _HAS_HINTED = True
        MediumSeverityWarning(
            f"YAML parsing is taking a long time.\n{HINT_LEAD_TEXT}Consider installing the `libyaml` package for faster parsing."
            f" See [link={URL.libyaml}]{URL.libyaml}[/link] for more information."
        ).print_warning()

    return result


# Spaces are allowed, but we replace them as well
_ILLEGAL_CHARACTERS = re.compile(r"[<>:\"/\\|?*\s]")


def sanitize_filename(text: str) -> str:
    """Convert a string to be a valid filename by replacing illegal characters with underscores."""
    cleaned = _ILLEGAL_CHARACTERS.sub("_", text)
    # Replace multiple underscores with a single one
    return re.sub(r"_+", "_", cleaned)


@contextmanager
def tmp_build_directory() -> typing.Generator[Path, None, None]:
    build_dir = Path(tempfile.mkdtemp(prefix="build.", suffix=".tmp", dir=Path.cwd()))
    try:
        yield build_dir
    finally:
        safe_rmtree(build_dir)


def safe_read(file: Path | str, encoding: str | None = None) -> str:
    """Falls back on explicit using utf-8 if the default .read_text()"""
    encoding = encoding or CDFToml.load().cdf.file_encoding
    if isinstance(file, str):
        return file
    try:
        return file.read_text(encoding=encoding)
    except UnicodeDecodeError:
        # If we set the encoding, we try the system default. If we tried the system default, we try utf-8.
        backup_encoding: str | None = None
        if encoding is None:
            backup_encoding = "utf-8"
        # On Windows, we may have issues as the encoding is not always utf-8
        try:
            return file.read_text(encoding=backup_encoding)
        except UnicodeDecodeError:
            raise


def safe_write(file: Path, content: str, encoding: str | None = None) -> None:
    """Falls back on explicit using utf-8 if the default .write_text()"""
    try:
        file.write_text(content, encoding=encoding)
    except UnicodeEncodeError:
        # On Windows, we may have issues as the encoding is not always utf-8
        file.write_text(content, encoding="utf-8")


def yaml_safe_dump(data: Any, sort_keys: bool = False, indent: int | None = None) -> str:
    return yaml.safe_dump(data, sort_keys=sort_keys, allow_unicode=True, indent=indent)


def quote_int_value_by_key_in_yaml(content: str, key: str) -> str:
    """Quote a value in a yaml string"""
    # This pattern will match the key if it is not already quoted
    pattern = rf"^(\s*-?\s*)?{key}:\s*(?!.*['\":])([\d_]+)$"
    replacement = rf'\1{key}: "\2"'

    return re.sub(pattern, replacement, content, flags=re.MULTILINE)


def stringify_value_by_key_in_yaml(content: str, key: str) -> str:
    """Quote a value in a yaml string"""
    pattern = rf"^{key}:\s*$"
    replacement = rf"{key}: |"
    return re.sub(pattern, replacement, content, flags=re.MULTILINE)


@dataclass(frozen=True)
class YAMLComment:
    """This represents a comment in a YAML file. It can be either above or after a variable."""

    above: list[str] = field(default_factory=list)
    after: list[str] = field(default_factory=list)

    @property
    def comment(self) -> str:
        return "\n".join(self.above) + "\n" + "\n".join(self.after)


T_Key = TypeVar("T_Key")
T_Value = TypeVar("T_Value")


class YAMLWithComments(UserDict[T_Key, T_Value]):
    @staticmethod
    def _extract_comments(raw_file: str, key_prefix: tuple[str, ...] = tuple()) -> dict[tuple[str, ...], YAMLComment]:
        """Extract comments from a raw file and return a dictionary with the comments."""
        comments: dict[tuple[str, ...], YAMLComment] = defaultdict(YAMLComment)
        position: Literal["above", "after"]
        init_value: object = object()
        variable: str | None | object = init_value
        last_comments: list[str] = []
        last_variable: str | None = None
        last_leading_spaces = 0
        parent_variables: list[str] = []
        indent: int | None = None
        for line in raw_file.splitlines():
            if ":" in line:
                # Is variable definition
                leading_spaces = len(line) - len(line.lstrip())
                variable = str(line.split(":", maxsplit=1)[0].strip())
                if leading_spaces > last_leading_spaces and last_variable:
                    parent_variables.append(last_variable)
                    if indent is None:
                        # Automatically indent based on the first variable
                        indent = leading_spaces
                elif leading_spaces < last_leading_spaces and parent_variables:
                    parent_variables = parent_variables[: -((last_leading_spaces - leading_spaces) // (indent or 2))]

                if last_comments:
                    comments[(*key_prefix, *parent_variables, variable)].above.extend(last_comments)
                    last_comments.clear()

                last_variable = variable
                last_leading_spaces = leading_spaces

            if "#" in line:
                # Potentially has comment.
                before, comment = str(line).rsplit("#", maxsplit=1)
                position = "after" if ":" in before else "above"
                if position == "after" and (before.count('"') % 2 == 1 or before.count("'") % 2 == 1):
                    # The comment is inside a string
                    continue
                # This is a new comment.
                if (position == "after" or variable is None) and variable is not init_value:
                    key = (*key_prefix, *parent_variables, *((variable and [variable]) or []))  # type: ignore[misc]
                    if position == "after":
                        comments[key].after.append(comment.strip())
                    else:
                        comments[key].above.append(comment.strip())
                else:
                    last_comments.append(comment.strip())

        return dict(comments)

    def _dump_yaml_with_comments(self, indent_size: int = 2, newline_after_indent_reduction: bool = False) -> str:
        """Dump a config dictionary to a yaml string"""
        config = self.dump()
        dumped = yaml.dump(config, sort_keys=False, indent=indent_size, width=float("inf"))
        out_lines = []
        if comments := self._get_comment(tuple()):
            for comment in comments.above:
                out_lines.append(f"# {comment}")
        last_indent = 0
        last_variable: str | None = None
        path: tuple[str, ...] = tuple()
        for line in dumped.splitlines():
            indent = len(line) - len(line.lstrip())
            if last_indent < indent:
                if last_variable is None:
                    raise ValueError("Unexpected state of last_variable being None")
                path = (*path, last_variable)
            elif last_indent > indent:
                if newline_after_indent_reduction:
                    # Adding some extra space between modules
                    out_lines.append("")
                indent_reduction_steps = (last_indent - indent) // indent_size
                path = path[:-indent_reduction_steps]

            variable = line.split(":", maxsplit=1)[0].strip()
            if comments := self._get_comment((*path, variable)):
                for line_comment in comments.above:
                    out_lines.append(f"{' ' * indent}# {line_comment}")
                if after := comments.after:
                    line = f"{line} # {after[0]}"

            out_lines.append(line)
            last_indent = indent
            last_variable = variable
        out_lines.append("")
        return "\n".join(out_lines)

    @abstractmethod
    def dump(self) -> dict[str, Any]: ...

    @abstractmethod
    def _get_comment(self, key: tuple[str, ...]) -> YAMLComment | None: ...

    # This is to get better type hints in the IDE
    def items(self) -> ItemsView[T_Key, T_Value]:
        return super().items()

    def keys(self) -> KeysView[T_Key]:
        return super().keys()

    def values(self) -> ValuesView[T_Value]:
        return super().values()


def remove_trailing_newline(content: str) -> str:
    """Remove the trailing newline character from a string"""
    while content.endswith("\n"):
        content = content[:-1]
    return content


def read_csv(
    path: Path | typing.TextIO,
    parse_dates: bool | None = None,
    index_col: Hashable | None = None,
    dtype: Any | None = None,
) -> pd.DataFrame:
    """Reads CSV

    Args:
        path (Path): Path to the CSV file.
        parse_dates (bool, optional): Whether to parse dates. Defaults to None.
        index_col (Hashable, optional): Index column. Defaults to None.
        dtype (Any, optional): Data types. Defaults to None

    Returns:
        pd.DataFrame: DataFrame with the CSV data.
    """
    return pd.read_csv(path, parse_dates=parse_dates, index_col=index_col, dtype=dtype)


def _handle_remove_readonly(func: Any, path: Any, exc: Any) -> None:
    excvalue = exc[1]
    if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
        # Typically on Windows, if the file is read-only, first remove the read-only attribute
        # https://stackoverflow.com/questions/1213706/what-user-do-python-scripts-run-as-in-windows
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


def safe_rmtree(path: Path) -> None:
    try:
        shutil.rmtree(path, ignore_errors=False, onerror=_handle_remove_readonly)
    except PermissionError:
        if path.is_dir():
            name = "directory"
        elif path.is_file():
            name = "file"
        else:
            name = "path"
        MediumSeverityWarning(
            f"Failed to remove {name} {path.as_posix()}. You may need to remove it manually."
        ).print_warning()


def get_table_columns(table: Path) -> list[str]:
    """Get the columns of a table

    Args:
        table (Path): Path to the table

    Returns:
        list[str]: List of columns
    """
    if table.suffix == ".csv":
        return read_csv(table).columns.tolist()
    elif table.suffix == ".parquet":
        return pd.read_parquet(table).columns.tolist()
    else:
        raise ToolkitValueError(f"The file {table.name} is not a supported table format (csv, parquet)")


@contextmanager
def create_temporary_zip(directory: Path, zipname: str) -> typing.Generator[Path, None, None]:
    """
    Create a temporary zip file from a directory.

    Args:
        directory: The directory to zip.
        zipname: The name of the zip file.

    Returns:
        A context manager that yields the path to the zip file.

    """
    with tempfile.TemporaryDirectory() as tmpdir:
        zip_path = Path(tmpdir, zipname)
        with ZipFile(zip_path, "w", strict_timestamps=False) as zf:
            for root, _, files in os.walk(directory):
                root_path = Path(root)
                arc_root = root_path.relative_to(directory)

                # Add directory entry to preserve directory structure, including empty ones.
                zf.write(root_path, arcname=str(arc_root))

                for filename in files:
                    file_path = root_path / filename
                    arcname = file_path.relative_to(directory)
                    zf.write(file_path, arcname)
        yield zip_path



from pydantic import Field

from cognite_toolkit._cdf_tk.constants import (
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    SPACE_FORMAT_PATTERN,
)

from .base import ToolkitResource
from .view_field_definitions import ViewReference


class DataModelYAML(ToolkitResource):
    """Cognite Data Model resource.

    Data models group and structure views into reusable collections.
    A data model contains a set of views where the node types can
    refer to each other with direct relations and edges.
    """

    space: str = Field(
        description="The workspace for the data model, a unique identifier for the space.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="Human readable name for the data model.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the data model.",
        max_length=1024,
    )
    views: list[ViewReference] | None = Field(
        description="List of views included in this data model.",
        default=None,
    )



from datetime import datetime

from cognite.client.data_classes import TimeSeries, TimeSeriesWrite
from cognite.client.data_classes.data_modeling import NodeApplyResultList
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteTimeSeriesApply
from cognite.client.utils._time import datetime_to_ms

from cognite_toolkit._cdf_tk.client import ToolkitClient


class TestExtendedTimeSeriesAPI:
    def test_set_pending_instance_id(self, dev_cluster_client: ToolkitClient, dev_space: str) -> None:
        """Happy path for setting a pending instance ID on a time series.

        1. Create asset-centric time series.
        2. Insert data points.
        3. Set pending instance ID.
        4. Create a CogniteTimeSeries
        5. Retrieve data points using the Node ID.
        6. Insert more data points using the Node ID.
        7. Retrieve data points again using the external ID of the original asset-centric time series.
        """
        client = dev_cluster_client
        ts = TimeSeriesWrite(
            external_id="ts_toolkit_integration_test_happy_path",
            name="Toolkit Integration Test Happy Path",
            is_step=False,
            is_string=False,
        )
        cognite_ts = CogniteTimeSeriesApply(
            space=dev_space,
            external_id=ts.external_id,
            is_step=False,
            time_series_type="numeric",
            name="Toolkit Integration Test Happy Path",
        )
        datapoints = [
            {"timestamp": datetime_to_ms(datetime(2020, 1, 1, 0, 0, 0)), "value": 1.0},
            {"timestamp": datetime_to_ms(datetime(2020, 1, 2, 0, 0, 0)), "value": 2.0},
            {"timestamp": datetime_to_ms(datetime(2020, 1, 3, 0, 0, 0)), "value": 3.0},
        ]
        more_datapoints = [
            {"timestamp": datetime_to_ms(datetime(2020, 1, 4, 0, 0, 0)), "value": 4.0},
            {"timestamp": datetime_to_ms(datetime(2020, 1, 5, 0, 0, 0)), "value": 5.0},
        ]
        created: TimeSeries | None = None
        created_dm: NodeApplyResultList | None = None
        try:
            created = client.time_series.create(ts)
            client.time_series.data.insert(datapoints, external_id=ts.external_id)
            updated = client.time_series.set_pending_ids(cognite_ts.as_id(), external_id=ts.external_id)
            assert updated.pending_instance_id == cognite_ts.as_id()

            created_dm = client.data_modeling.instances.apply(cognite_ts).nodes

            retrieve_datapoints = client.time_series.data.retrieve(instance_id=cognite_ts.as_id())
            assert len(retrieve_datapoints) == len(datapoints)
            assert retrieve_datapoints.dump()["datapoints"] == datapoints
            client.time_series.data.insert(more_datapoints, instance_id=cognite_ts.as_id())

            retrieve_datapoints2 = client.time_series.data.retrieve(external_id=ts.external_id)
            assert len(retrieve_datapoints2) == len(datapoints) + len(more_datapoints)
            assert retrieve_datapoints2.dump()["datapoints"] == datapoints + more_datapoints

            retrieved_ts = client.time_series.retrieve(external_id=ts.external_id)
            assert retrieved_ts.instance_id == cognite_ts.as_id()
            listed = client.time_series.list(external_id_prefix=ts.external_id)
            assert len(listed) == 1
            assert listed[0].external_id == ts.external_id
        finally:
            if created is not None and created_dm is None:
                client.time_series.delete(external_id=ts.external_id)
            if created_dm is not None:
                # This will delete the CogniteTimeSeries and the asset-centric time series
                client.data_modeling.instances.delete(cognite_ts.as_id())

    def test_unlink_instance_ids(self, dev_cluster_client: ToolkitClient, space: str) -> None:
        client = dev_cluster_client
        ts = TimeSeriesWrite(
            external_id="ts_toolkit_integration_test_unlink",
            name="Toolkit Integration Test Unlink",
            is_step=False,
            is_string=False,
        )
        cognite_ts = CogniteTimeSeriesApply(
            space=space,
            external_id=ts.external_id,
            is_step=False,
            time_series_type="numeric",
            name="Toolkit Integration Test Unlink",
        )
        created: TimeSeries | None = None
        created_dm: NodeApplyResultList | None = None
        try:
            created = client.time_series.create(ts)
            updated = client.time_series.set_pending_ids(cognite_ts.as_id(), external_id=ts.external_id)
            assert updated.pending_instance_id == cognite_ts.as_id()

            created_dm = client.data_modeling.instances.apply(cognite_ts).nodes

            retrieved_ts = client.time_series.retrieve(instance_id=cognite_ts.as_id())
            assert retrieved_ts.id == created.id

            unlinked = client.time_series.unlink_instance_ids(id=created.id)
            assert unlinked.id == created.id

            client.data_modeling.instances.delete(cognite_ts.as_id())
            created_dm = None

            # Still existing time series.
            retrieved_ts = client.time_series.retrieve(external_id=ts.external_id)
            assert retrieved_ts is not None
            assert retrieved_ts.id == created.id
        finally:
            if created is not None and created_dm is None:
                client.time_series.delete(external_id=ts.external_id)
            if created_dm is not None:
                client.data_modeling.instances.delete(cognite_ts.as_id())



import platform
import random
import sys

# This is needed as we run tests for two different versions of Python in parallel.
# The platform.system() is not used, but is here in case we start testing on Windows as well.
# The random number is to avoid conflicts when running tests in parallel (for example, two PRs).
# CAUTION: This should only be used for ephemeral test resources that are cleaned up after the test run. It should
# not be used for resources that are expected to persist beyond the test run. For example, a data modeling space,
# that is used to host other resources, should not use this constant as part of its name.
RUN_UNIQUE_ID = (
    f"{platform.system()}_{sys.version_info.major}_{sys.version_info.minor}_{random.randint(0, 100_000_000)!s}"
)

# These constants are used to set up a raw database and tables with transformations and datasets
# to populate the asset-centric resources in CDF.
ASSET_TABLE = "toolkit_aggregators_test_table_assets"
EVENT_TABLE = "toolkit_aggregators_test_table_events"
FILE_TABLE = "toolkit_aggregators_test_table_files"
TIMESERIES_TABLE = "toolkit_aggregators_test_table_time_series"
SEQUENCE_TABLE = "toolkit_aggregators_test_table_sequences"

ASSET_TRANSFORMATION = "toolkit_aggregators_test_asset_transformation"
EVENT_TRANSFORMATION = "toolkit_aggregators_test_event_transformation"
FILE_TRANSFORMATION = "toolkit_aggregators_test_file_transformation"
TIMESERIES_TRANSFORMATION = "toolkit_aggregators_test_timeseries_transformation"
SEQUENCE_TRANSFORMATION = "toolkit_aggregators_test_sequence_transformation"

ASSET_DATASET = "toolkit_aggregators_test_dataset_1"
EVENT_DATASET = "toolkit_aggregators_test_dataset_1"
FILE_DATASET = "toolkit_aggregators_test_dataset_2"
TIMESERIES_DATASET = "toolkit_aggregators_test_dataset_2"
SEQUENCE_DATASET = "toolkit_aggregators_test_dataset_2"

ASSET_COUNT = 6
EVENT_COUNT = 10
FILE_COUNT = 3
TIMESERIES_COUNT = 20
SEQUENCE_COUNT = 2



from pydantic import Field

from cognite_toolkit._cdf_tk.resource_classes.base import ToolkitResource


class RobotLocationYAML(ToolkitResource):
    name: str = Field(description="Location name.")
    external_id: str = Field(description="Location external id. Must be unique for the resource type.", max_length=256)
    description: str | None = Field(
        default=None,
        description="Description of Location. Textual description of the Location.",
    )



from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest

from cognite_toolkit._cdf_tk.resource_classes.relationship import RelationshipYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_relationship_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "myRelationship", "targetExternalId": "target", "sourceExternalId": "source"},
        {"Missing required field: 'sourceType'", "Missing required field: 'targetType'"},
        id="Missing required fields",
    )
    yield pytest.param(
        {"externalId": "equipment:pump", "dataSetId": 123},
        {
            "Missing required field: 'sourceExternalId'",
            "Missing required field: 'sourceType'",
            "Missing required field: 'targetExternalId'",
            "Missing required field: 'targetType'",
            "Unused field: 'dataSetId'",
        },
        id="Unused field: dataSetId and missing name",
    )


class TestRelationshipYAML:
    @pytest.mark.parametrize("data", list(find_resources("Relationship")))
    def test_load_valid_relationship(self, data: dict[str, Any]) -> None:
        loaded = RelationshipYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_relationship_test_cases()))
    def test_invalid_relationship_error_messages(self, data: dict[str, Any], expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, RelationshipYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class FunctionScheduleID:
    function_external_id: str
    name: str

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "functionExternalId" if camel_case else "function_external_id": self.function_external_id,
            "name": self.name,
        }



from rich import print

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands.deploy import DeployCommand
from cognite_toolkit._cdf_tk.cruds import (
    ContainerCRUD,
    DataModelCRUD,
    ResourceCRUD,
    ResourceViewMappingCRUD,
    ResourceWorker,
    SpaceCRUD,
    ViewCRUD,
)
from cognite_toolkit._cdf_tk.data_classes import DeployResults
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning

from .data_model import COGNITE_MIGRATION_MODEL, CONTAINERS, MODEL_ID, SPACE, VIEWS
from .default_mappings import create_default_mappings


class MigrationPrepareCommand(ToolkitCommand):
    def deploy_cognite_migration(self, client: ToolkitClient, dry_run: bool, verbose: bool = False) -> DeployResults:
        """Deploys the Cognite Migration Data Model"""

        deploy_cmd = DeployCommand(self.print_warning, silent=self.silent)
        deploy_cmd.tracker = self.tracker

        verb = "Would deploy" if dry_run else "Deploying"
        print(f"{verb} {MODEL_ID!r}")
        results = DeployResults([], "deploy", dry_run=dry_run)
        crud_cls: type[ResourceCRUD]
        for crud_cls, resource_list in [  # type: ignore[assignment]
            (SpaceCRUD, [SPACE]),
            (ContainerCRUD, CONTAINERS),
            (ViewCRUD, VIEWS),
            (DataModelCRUD, [COGNITE_MIGRATION_MODEL]),
            (ResourceViewMappingCRUD, create_default_mappings()),
        ]:
            crud = crud_cls.create_loader(client)
            if warning := crud.prerequisite_warning():
                self.warn(HighSeverityWarning(warning))
                continue
            worker = ResourceWorker(crud, "deploy")
            # MyPy does not understand that `loader` has a `get_id` method.
            dump_arg = {"context": "local"} if crud_cls is ResourceViewMappingCRUD else {}
            local_by_id = {crud.get_id(item): (item.dump(**dump_arg), item) for item in resource_list}  # type: ignore[attr-defined]
            worker.validate_access(local_by_id, is_dry_run=dry_run)
            cdf_resources = crud.retrieve(list(local_by_id.keys()))
            resources = worker.categorize_resources(local_by_id, cdf_resources, False, verbose)

            if dry_run:
                result = deploy_cmd.dry_run_deploy(resources, crud, False, False)
            else:
                result = deploy_cmd.actual_deploy(resources, crud)
            if result:
                results[result.name] = result
        if results.has_counts:
            print(results.counts_table())
        return results



"""This is a small CLI used to develop Toolkit.

It is not used in production and is only intended to be used for development purposes.
"""

import itertools
import re
from pathlib import Path
from typing import Literal, get_args

import marko
import marko.block
import marko.element
import marko.inline
import typer
from packaging.version import Version, parse
from rich import print

REPO_ROOT = Path(__file__).parent
CHANGELOG = REPO_ROOT / "CHANGELOG.cdf-tk.md"
TEMPLATE_CHANGELOG = REPO_ROOT / "CHANGELOG.templates.md"
TBD_HEADING = "## TBD"
IMAGE_NAME = "cognite/toolkit"
CDF_TOML = REPO_ROOT / "cdf.toml"

VALID_CHANGELOG_HEADERS = {"Added", "Changed", "Removed", "Fixed", "Improved"}
BUMP_OPTIONS = Literal["major", "minor", "patch", "skip"]
VALID_BUMP_OPTIONS = get_args(BUMP_OPTIONS)
LAST_GIT_MESSAGE_FILE = REPO_ROOT / "last_git_message.txt"
CHANGELOG_ENTRY_FILE = REPO_ROOT / "last_changelog_entry.md"
LAST_VERSION = REPO_ROOT / "last_version.txt"
VERSION_PLACEHOLDER = "0.0.0"
VERSION_FILES = [
    REPO_ROOT / "pyproject.toml",
    REPO_ROOT / "cognite_toolkit" / "_version.py",
    REPO_ROOT / "cdf.toml",
    *(REPO_ROOT / "tests" / "data").rglob("cdf.toml"),
    *(REPO_ROOT / "tests" / "data").rglob("_build_environment.yaml"),
    *(REPO_ROOT / "cognite_toolkit").rglob("cdf.toml"),
    REPO_ROOT / "cognite_toolkit" / "_repo_files" / "GitHub" / ".github" / "workflows" / "dry-run.yaml",
    REPO_ROOT / "cognite_toolkit" / "_repo_files" / "GitHub" / ".github" / "workflows" / "deploy.yaml",
]

DOCKER_IMAGE_FILES = [
    *(REPO_ROOT / "cognite_toolkit" / "_repo_files").rglob("*.yml"),
    *(REPO_ROOT / "cognite_toolkit" / "_repo_files").rglob("*.yaml"),
]

app = typer.Typer(
    add_completion=False,
    help=__doc__,
    pretty_exceptions_short=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=False,
)


@app.command()
def bump(verbose: bool = False) -> None:
    last_version_str = LAST_VERSION.read_text().strip().removeprefix("v")
    try:
        last_version = parse(last_version_str)
    except ValueError:
        print(f"Invalid last version: {last_version_str}")
        raise SystemExit(1)

    changelog_items, _ = _read_last_commit_message()
    version_bump = _get_change(changelog_items)
    if version_bump == "skip":
        print("No changes to release.")
        return
    if version_bump == "major":
        new_version = Version(f"{last_version.major + 1}.0.0")
    elif version_bump == "minor":
        new_version = Version(f"{last_version.major}.{last_version.minor + 1}.0")
    elif version_bump == "patch":
        new_version = Version(f"{last_version.major}.{last_version.minor}.{last_version.micro + 1}")
    else:
        raise typer.BadParameter("You must specify one of major, minor, patch, alpha, or beta.")

    for file in VERSION_FILES:
        file.write_text(file.read_text().replace(str(VERSION_PLACEHOLDER), str(new_version), 1))
        if verbose:
            typer.echo(f"Bumped version from {last_version} to {new_version} in {file}.")
    for file in DOCKER_IMAGE_FILES:
        file.write_text(file.read_text().replace(f"{IMAGE_NAME}:{last_version!s}", f"{IMAGE_NAME}:{new_version!s}", 1))
        if verbose:
            typer.echo(f"Bumped version from {last_version} to {new_version} in {file}.")

    typer.echo(f"Bumped version from {last_version} to {new_version} in {len(VERSION_FILES)} files.")


@app.command("alpha")
def set_alpha(off: bool = False) -> None:
    if not off:
        return
    is_feature_flag = False
    new_lines = []
    for line in CDF_TOML.read_text().splitlines():
        if header_match := re.match(r"\[(\w+)\]", line.strip()):
            header = header_match.group(1)
            print(header)
            if header == "alpha_flags":
                is_feature_flag = True
            else:
                is_feature_flag = False
        if is_feature_flag:
            line = line.replace("true", "false")
        new_lines.append(line)

    CDF_TOML.write_text("\n".join(new_lines) + "\n")


@app.command("changelog")
def create_changelog_entry() -> None:
    changelog_items, changelog_text = _read_last_commit_message()
    version_bump = _get_change(changelog_items)
    if version_bump == "skip":
        print("No changes to release.")
        return
    if not changelog_items[1:]:
        print(f"Trying to {version_bump} bump but no changes found in the changelog.")
        raise SystemExit(1)
    if not _is_header(changelog_items[1], level=2, text="cdf"):
        print("The first header in the changelog must be '## cdf'.")
        raise SystemExit(1)
    cdf_entries = list(itertools.takewhile(lambda x: not _is_header(x, level=2), changelog_items[2:]))
    _validate_entries(cdf_entries, "cdf")
    no = next((no for no, item in enumerate(changelog_items) if _is_header(item, level=2, text="templates")), None)
    if no is None:
        print("No '## templates' section found in the changelog.")
        raise SystemExit(1)
    if not changelog_items[no + 1 :]:
        print("No template entries found in the changelog.")
        raise SystemExit(1)
    template_entries = list(changelog_items[no + 1 :])
    _validate_entries(template_entries, "templates")

    changelog_entry = changelog_text.split("## cdf")[1]
    CHANGELOG_ENTRY_FILE.write_text(f"## cdf {changelog_entry}", encoding="utf-8")
    print(f"Changelog entry written to {CHANGELOG_ENTRY_FILE}.")


def _read_last_commit_message() -> tuple[list[marko.element.Element], str]:
    last_git_message = LAST_GIT_MESSAGE_FILE.read_text()
    if "This PR was generated by [Mend Renovate]" in last_git_message:
        print("Skipping Renovate PR.")
        changelog_text = "- [ ] Patch\n- [ ] Minor\n- [x] Skip\n"
    elif "## Changelog" not in last_git_message:
        print("No changelog entry found in the last commit message.")
        raise SystemExit(1)
    else:
        changelog_text = last_git_message.split("## Changelog")[1].strip()

    if "-----" in changelog_text:
        # Co-authors section
        changelog_text = changelog_text.split("-----")[0].strip()

    changelog_items = [
        item for item in marko.parse(changelog_text).children if not isinstance(item, marko.block.BlankLine)
    ]
    if not changelog_items:
        print("No changelog items found in the last commit message.")
        raise SystemExit(1)

    return changelog_items, changelog_text


def _is_header(item: marko.element.Element, level: int, text: str | None = None):
    if not (
        isinstance(item, marko.block.Heading)
        and item.level == level
        and isinstance(item.children[0], marko.inline.RawText)
    ):
        return False
    return text is None or item.children[0].children == text


def _get_change(changelog_items: list[marko.element.Element]) -> Literal["major", "minor", "patch", "skip"]:
    item = changelog_items[0]
    if not isinstance(item, marko.block.List):
        print("The first item in the changelog must be a list with the type of change.")
        raise SystemExit(1)
    selected: list[Literal["major", "minor", "patch", "skip"]] = []
    for child in item.children:
        if not isinstance(child, marko.block.ListItem):
            print(f"Unexpected item in changelog: {child}")
            raise SystemExit(1)
        if not isinstance(child.children[0], marko.block.Paragraph):
            print(f"Unexpected item in changelog: {child.children[0]}")
            raise SystemExit(1)
        if not isinstance(child.children[0].children[0], marko.inline.RawText):
            print(f"Unexpected item in changelog: {child.children[0].children[0]}")
            raise SystemExit(1)
        list_text = child.children[0].children[0].children
        if list_text.startswith("[ ]"):
            continue
        elif list_text.lower().startswith("[x]"):
            change_type = list_text.removeprefix("[x]").removeprefix("[X]").strip()
            if change_type.casefold() not in VALID_BUMP_OPTIONS:
                print(f"Unexpected change type in changelog: {change_type}")
                raise SystemExit(1)
            selected.append(change_type.casefold())
        else:
            print(f"Unexpected item in changelog: {list_text}")
            raise SystemExit(1)

    if len(selected) > 1:
        print("You can only select one type of change.")
        raise SystemExit(1)
    if not selected:
        print("You must select a type of change.")
        raise SystemExit(1)
    return selected[0]


def _validate_entries(items: list[marko.element.Element], section: str) -> None:
    seen_headers: set[str] = set()
    if not items:
        print(f"No entries found in the {section} section of the changelog.")
        raise SystemExit(1)
    if (
        isinstance(items[0], marko.block.Paragraph)
        and isinstance(items[0].children[0], marko.inline.RawText)
        and "no changes" in items[0].children[0].children.casefold()
    ):
        return
    last_header: str = ""
    for item in items:
        if isinstance(item, marko.block.Heading):
            if last_header:
                print(f"Expected a list of changes after the {last_header} header.")
                raise SystemExit(1)
            elif item.level != 3:
                print(f"Unexpected header level in changelog: {item}. Should be level 3.")
                raise SystemExit(1)
            elif not isinstance(item.children[0], marko.inline.RawText):
                print(f"Unexpected header in changelog: {item}.")
                raise SystemExit(1)
            header_text = item.children[0].children
            if header_text not in VALID_CHANGELOG_HEADERS:
                print(f"Unexpected header in changelog: {VALID_CHANGELOG_HEADERS}.")
                raise SystemExit(1)
            if header_text in seen_headers:
                print(f"Duplicate header in changelog: {header_text}.")
                raise SystemExit(1)
            seen_headers.add(header_text)
            last_header = header_text
        elif isinstance(item, marko.block.List):
            if not last_header:
                print("Expected a header before the list of changes.")
                raise SystemExit(1)
            last_header = ""
        else:
            print(f"Unexpected item in changelog: {item}.")
            raise SystemExit(1)


# This is just for demo purposes, to test the secret plugin in the Toolkit CLI
import_app = typer.Typer(
    pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False
)


@import_app.command("cdf")
def cdf(
    ctx: typer.Context,
) -> None:
    """Import resources into Cognite Data Fusion."""
    print("Ran CDF Import Command")


CDF_TK_PLUGIN = {
    "bump": app,
    "import": import_app,
}

if __name__ == "__main__":
    app()



from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import AuthCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class AuthApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command()(self.init)
        self.command()(self.verify)

    def main(self, ctx: typer.Context) -> None:
        """Commands to auth setup"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf auth --help[/] for more information.")

    def init(
        self,
        no_verify: Annotated[
            bool,
            typer.Option(
                "--no-verify",
                "-nv",
                help="Whether to skip the verification of the capabilities after the initialization.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="If you verify, and you pass this flag no changes to CDF will be made.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Sets the OIDC parameters required to authenticate and authorize the Cognite Toolkit in Cognite Data Fusion.

        This will prompt the user to log in and optionally store the credentials in a .env file.

        Needed capabilities for bootstrapping:
        "projectsAcl": ["LIST", "READ"],
        "groupsAcl": ["LIST", "READ", "CREATE", "UPDATE", "DELETE"]
        """
        cmd = AuthCommand()
        cmd.run(
            lambda: cmd.init(
                no_verify=no_verify,
                dry_run=dry_run,
            )
        )

    def verify(
        self,
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run. This means that no changes to CDF will be made.",
            ),
        ] = False,
        no_prompt: Annotated[
            bool,
            typer.Option(
                "--no-prompt",
                "-np",
                help="Whether to skip the prompt to continue. This is useful for CI/CD pipelines."
                "If you include this flag, the execution will stop if the user or service principal does not have the required capabilities.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Verify that the current user or service principal has the required capabilities to run the CDF Toolkit commands."""
        cmd = AuthCommand()
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd.run(
            lambda: cmd.verify(
                client,
                dry_run=dry_run,
                no_prompt=no_prompt,
            )
        )



from abc import ABC
from typing import Literal

from ._base import DataSelector


class ChartSelector(DataSelector, ABC):
    kind: Literal["Charts"] = "Charts"


class ChartOwnerSelector(ChartSelector):
    type: Literal["chartOwner"] = "chartOwner"
    owner_id: str

    @property
    def group(self) -> str:
        return "Charts"

    def __str__(self) -> str:
        return self.owner_id


class AllChartsSelector(ChartSelector):
    type: Literal["allCharts"] = "allCharts"

    @property
    def group(self) -> str:
        return "Charts"

    def __str__(self) -> str:
        return "all"



from __future__ import annotations

import json
from pathlib import Path
from typing import Union
from unittest.mock import MagicMock

import pytest
import yaml
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList, CogniteResponse
from cognite.client.data_classes.capabilities import (
    AllProjectsScope,
    AllScope,
    AppConfigAcl,
    AppConfigScope,
    AssetsAcl,
    Capability,
    ProjectCapability,
    ProjectCapabilityList,
    RelationshipsAcl,
)
from cognite.client.data_classes.iam import Group, GroupList, ProjectSpec, TokenInspection

from cognite_toolkit._cdf_tk.client.data_classes.project import ProjectStatusList
from cognite_toolkit._cdf_tk.commands import AuthCommand
from cognite_toolkit._cdf_tk.exceptions import AuthorizationError
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    LowSeverityWarning,
    MissingCapabilityWarning,
    ToolkitWarning,
    WarningList,
)
from tests.data import AUTH_DATA
from tests.test_unit.conftest import ApprovalToolkitClient


@pytest.fixture
def cdf_resources() -> dict[type[CogniteResource] | type[CogniteResponse], CogniteResource | CogniteResourceList]:
    # Load the rw-group and add it to the mock client as an existing resource
    group = Group.load(yaml.safe_load((AUTH_DATA / "rw-group.yaml").read_text()))
    project_capabilities: ProjectCapabilityList = ProjectCapabilityList([])
    for cap in group.capabilities:
        project_capabilities.append(ProjectCapability(capability=cap, project_scope=AllProjectsScope()))
    inspect_result = TokenInspection(
        subject="test@test.com",
        projects=[ProjectSpec("pytest-project", groups=[1234567890])],
        capabilities=ProjectCapabilityList(project_capabilities),
    )
    return {
        TokenInspection: inspect_result,
        Group: GroupList([group]),
        # NOTE! If you add more resources to be pre-loaded in the CDF project, add them here.
    }


@pytest.fixture
def auth_cognite_approval_client(
    toolkit_client_approval: ApprovalToolkitClient,
) -> ApprovalToolkitClient:
    # Mock the get call to return the project info
    def mock_get_json(*args, **kwargs):
        mock = MagicMock()
        mock.json.return_value = json.loads(Path(AUTH_DATA / "project_info.json").read_text())
        return mock

    # Set the mock get call to return the project info
    toolkit_client_approval.client.get.side_effect = mock_get_json
    # Returning empty list means no capabilities are missing
    toolkit_client_approval.client.iam.verify_capabilities.return_value = []
    return toolkit_client_approval


@pytest.mark.skip("Needs to be rewritten to support interactive")
class TestAuthCommand:
    def test_auth_verify_happy_path(
        self,
        auth_cognite_approval_client: ApprovalToolkitClient,
        cdf_resources: dict[type[CogniteResource], CogniteResource | CogniteResourceList],
    ):
        # First, add the pre-loaded data to the approval_client
        for resource, data in cdf_resources.items():
            auth_cognite_approval_client.append(resource, data)
        # Then make sure that the CogniteClient used is the one mocked by
        # the approval_client
        cmd = AuthCommand(print_warning=False)

        cmd.verify(auth_cognite_approval_client.mock_client, False, no_prompt=True)

        assert list(cmd.warning_list) == []

    def test_auth_verify_wrong_capabilities(
        self,
        auth_cognite_approval_client: ApprovalToolkitClient,
        cdf_resources: dict[type[CogniteResource], CogniteResource | CogniteResourceList],
    ):
        expected_warnings = WarningList[ToolkitWarning]()
        # Remove the last 3 capabilities from the inspect result to make a test case
        # with wrong capabilities in the current CDF group.
        for _ in range(1, 4):
            capability = cdf_resources[TokenInspection].capabilities.pop()
            for cap in capability.capability.as_tuples():
                expected_warnings.append(MissingCapabilityWarning(str(Capability.from_tuple(cap))))
        # Add the pre-loaded data to the approval_client
        for resource, data in cdf_resources.items():
            auth_cognite_approval_client.append(resource, data)
        # Then make sure that the CogniteClient used is the one mocked by
        # the approval_client
        cmd = AuthCommand(print_warning=False)

        cmd.verify(auth_cognite_approval_client.mock_client, False, no_prompt=True)

        assert len(cmd.warning_list) == len(expected_warnings)
        assert set(cmd.warning_list) == set(expected_warnings)

    def test_auth_verify_two_groups(
        self,
        auth_cognite_approval_client: ApprovalToolkitClient,
        cdf_resources: dict[CogniteResource, Union[CogniteResource, CogniteResourceList]],
    ):
        # Add another group
        cdf_resources[Group].append(Group.load(yaml.safe_load((AUTH_DATA / "rw-group.yaml").read_text())))
        cdf_resources[Group][1].name = "2nd group"

        # Add the pre-loaded data to the approval_client
        for resource, data in cdf_resources.items():
            auth_cognite_approval_client.append(resource, data)
        # Then make sure that the CogniteClient used is the one mocked by
        # the approval_client
        cmd = AuthCommand(print_warning=False)
        cmd.verify(auth_cognite_approval_client.mock_client, False, no_prompt=True)

        assert len(cmd.warning_list) == 1
        assert set(cmd.warning_list) == {
            LowSeverityWarning(
                "This service principal/application gets its "
                "access rights from more than one CDF "
                "group.           This is not recommended. The "
                "group matching the group config file is "
                "marked in bold above if it is present."
            )
        }

    def test_auth_verify_no_capabilities(
        self,
        auth_cognite_approval_client: ApprovalToolkitClient,
        cdf_resources: dict[type[CogniteResource], Union[CogniteResource, CogniteResourceList]],
    ):
        # Add the pre-loaded data to the approval_client
        for resource, data in cdf_resources.items():
            auth_cognite_approval_client.append(resource, data)

        def mock_verify_client(*args, **kwargs):
            raise Exception("No capabilities")

        cmd = AuthCommand(print_warning=False)
        with pytest.raises(AuthorizationError) as e:
            cmd.verify(auth_cognite_approval_client.mock_client, False, no_prompt=True)

        assert len(cmd.warning_list) == 1
        assert set(cmd.warning_list) == {
            HighSeverityWarning(
                "The service principal/application configured for this client "
                "does not have the basic group write access rights."
            )
        }
        assert str(e.value) == (
            "Unable to continue, the service principal/application configured for this "
            "client does not have the basic read group access rights."
        )


def test_get_capabilities_by_loader_hybrid_project(toolkit_client_approval: ApprovalToolkitClient):
    client = toolkit_client_approval.client
    client.project = MagicMock()
    client.project.status.return_value = ProjectStatusList._load(
        [
            {
                "urlName": client.config.project,
                "dataModelingStatus": "HYBRID",
            }
        ],
        cognite_client=client,
    )
    caps_hybrid, _ = AuthCommand._get_capabilities_by_loader(client)
    cap_types_hybrid = {type(c) for c in caps_hybrid}
    assert AssetsAcl in cap_types_hybrid
    assert RelationshipsAcl in cap_types_hybrid


def test_get_capabilities_by_loader_dm_only_project(toolkit_client_approval: ApprovalToolkitClient):
    client = toolkit_client_approval.client
    client.project = MagicMock()
    client.project.status.return_value = ProjectStatusList._load(
        [
            {
                "urlName": client.config.project,
                "dataModelingStatus": "DATA_MODELING_ONLY",
            }
        ],
        cognite_client=client,
    )
    caps_dm_only, _ = AuthCommand._get_capabilities_by_loader(client)
    cap_types_dm_only = {type(c) for c in caps_dm_only}
    assert AssetsAcl not in cap_types_dm_only
    assert RelationshipsAcl not in cap_types_dm_only


@pytest.mark.parametrize(
    "capability_list,expected_result",
    [
        pytest.param(
            [
                AppConfigAcl(
                    actions=[AppConfigAcl.Action.Read, AppConfigAcl.Action.Write], scope=AppConfigScope(apps=["SEARCH"])
                ),
                AppConfigAcl(actions=[AppConfigAcl.Action.Read], scope=AppConfigScope(apps=["SEARCH"])),
                AppConfigAcl(actions=[AppConfigAcl.Action.Write], scope=AppConfigScope(apps=["SEARCH"])),
                AssetsAcl(actions=[AssetsAcl.Action.Read, AssetsAcl.Action.Write], scope=AllScope()),
            ],
            [
                AppConfigAcl(
                    actions=[AppConfigAcl.Action.Read, AppConfigAcl.Action.Write], scope=AppConfigScope(apps=["SEARCH"])
                ),
                AssetsAcl(actions=[AssetsAcl.Action.Read, AssetsAcl.Action.Write], scope=AllScope()),
            ],
            id="Merge multiple capabilities",
        )
    ],
)
def test_merge_capabilities(capability_list: list[Capability], expected_result: list[Capability]):
    merged_capabilities = AuthCommand.merge_capabilities(capability_list)
    assert len(merged_capabilities) == len(expected_result)
    app_config_acl = next(c for c in merged_capabilities if type(c) is AppConfigAcl)
    assert app_config_acl is not None
    assert app_config_acl.scope == expected_result[0].scope
    assert sorted(app_config_acl.dump()["appConfigAcl"]["actions"]) == sorted(
        expected_result[0].dump()["appConfigAcl"]["actions"]
    )



from typing import Annotated, Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.constants import DM_EXTERNAL_ID_PATTERN, DM_VERSION_PATTERN, SPACE_FORMAT_PATTERN

from .base import BaseModelResource


class AgentToolDefinition(BaseModelResource):
    type: str
    name: str = Field(
        description="A name for the tool, unique within the agent.",
        min_length=1,
        max_length=64,
        pattern=r"^[^\x00]{1,64}$",
    )
    description: str = Field(
        description="A description of how the tool helps the language model understand when and how to use the tool.",
        min_length=10,
        max_length=1024,
    )


class AskDocument(AgentToolDefinition):
    type: Literal["askDocument"] = "askDocument"


class AgentDataModel(BaseModelResource):
    space: str = Field(
        description="The space the data model is in.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="The external ID of the data model.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="The version of the data model.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    view_external_ids: list[str] = Field(description="The views of the data model.", min_length=1, max_length=10)


class AgentInstanceSpacesDefinition(BaseModelResource):
    type: str


class AllInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["all"] = "all"


class ManualInstanceSpaces(AgentInstanceSpacesDefinition):
    type: Literal["manual"] = "manual"
    spaces: list[str]


AgentInstanceSpaces = Annotated[
    AllInstanceSpaces | ManualInstanceSpaces,
    Field(discriminator="type"),
]


class QueryKnowledgeGraphConfig(BaseModelResource):
    data_models: list[AgentDataModel] = Field(
        description="List of relevant data models.",
        min_length=1,
        max_length=80,
    )
    instance_spaces: AgentInstanceSpaces
    version: Literal["v1", "v2"] = Field(
        "v2",
        description="The version of the query generation strategy to use. "
        "A higher number does not necessarily mean a better query.",
    )


class QueryKnowledgeGraph(AgentToolDefinition):
    type: Literal["queryKnowledgeGraph"] = "queryKnowledgeGraph"
    configuration: QueryKnowledgeGraphConfig = Field(
        description="Configuration for the Query Knowledge Graph tool.",
    )


class QueryTimeSeriesDatapoints(AgentToolDefinition):
    type: Literal["queryTimeSeriesDatapoints"] = "queryTimeSeriesDatapoints"


class SummarizeDocument(AgentToolDefinition):
    type: Literal["summarizeDocument"] = "summarizeDocument"


AgentTool = Annotated[
    AskDocument | QueryKnowledgeGraph | QueryTimeSeriesDatapoints | SummarizeDocument,
    Field(discriminator="type"),
]



from collections.abc import Hashable
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Union

from cognite_toolkit._cdf_tk.tk_warnings.base import GeneralWarning, SeverityLevel, ToolkitWarning


@dataclass(frozen=True)
class UnexpectedFileLocationWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    filepath: str
    alternative: str

    def get_message(self) -> str:
        return f"{self.filepath!r} does not exist. Using {self.alternative!r} instead."


@dataclass(frozen=True)
class ToolkitBugWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    message: ClassVar[str] = "Please contact the toolkit maintainers with the error message and traceback:"
    header: str
    traceback: str

    def get_message(self) -> str:
        return " ".join((self.header, self.message, self.traceback))


@dataclass(frozen=True)
class IncorrectResourceWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: ClassVar[str] = "The resource not semantically correct:"
    location: str
    resource: str
    details: str | list[str] | None = None

    def get_message(self) -> str:
        extra_details = []
        if self.details:
            if isinstance(self.details, str):
                extra_details.append(self.details)
            else:
                extra_details.extend(self.details)
        return " ".join((self.location, self.message, self.resource, *extra_details))


@dataclass(frozen=True)
class LowSeverityWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message_raw: str

    def get_message(self) -> str:
        return self.message_raw


@dataclass(frozen=True)
class MediumSeverityWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.MEDIUM
    message_raw: str

    def get_message(self) -> str:
        return self.message_raw


@dataclass(frozen=True)
class HighSeverityWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    message_raw: str

    def get_message(self) -> str:
        return self.message_raw


@dataclass(frozen=True)
class ToolkitDependenciesIncludedWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: ClassVar[str] = "Operation may fail due to missing dependencies"
    dependencies: Union[None, str, list[str]]

    def get_message(self) -> str:
        if not self.dependencies:
            return f"{self.message}."
        dep = self.dependencies if isinstance(self.dependencies, list) else [self.dependencies]
        dep_str = ", ".join(f"{d!r}" for d in sorted(dep))
        return f"{self.message}: {dep_str}."


@dataclass(frozen=True)
class ToolkitNotSupportedWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: ClassVar[str] = "This feature is not supported"
    feature: str

    details: str | list[str] | None = None

    def get_message(self) -> str:
        extra_details = []
        if self.details:
            if isinstance(self.details, str):
                extra_details.append(self.details)
            else:
                extra_details.extend(self.details)
        return " ".join((self.message, self.feature, *extra_details))


@dataclass(frozen=True)
class MissingDependencyWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.HIGH
    message: ClassVar[str] = ""
    dependency_type: str
    identifier: Hashable
    required_by: set[tuple[Hashable, Path]]
    has_checked_cdf: bool = False

    def get_message(self) -> str:
        location = "local configurations"
        if self.has_checked_cdf:
            location = "in CDF and local configurations"
        msg = f"{self.dependency_type} {self.identifier!r} is missing {location} and is required by:"
        for identifier, path in self.required_by:
            msg += f"\n- {identifier!r} in {path}"
        return msg


@dataclass(frozen=True)
class MissingCapabilityWarning(GeneralWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.MEDIUM
    message: ClassVar[str] = "Missing capability:"
    capability: str

    def get_message(self) -> str:
        msg = f"The principal lacks the required access capability {self.capability} in the CDF project"
        return msg


@dataclass(frozen=True)
class ToolkitDeprecationWarning(ToolkitWarning, DeprecationWarning):
    message: ClassVar[str] = "The '{feature}' is deprecated and will be removed in a future version."

    feature: str
    alternative: str | None = None

    def get_message(self) -> str:
        msg = self.message.format(feature=self.feature)
        if self.alternative:
            msg += f"\nUse {self.alternative!r} instead."

        return msg


@dataclass(frozen=True)
class HTTPWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.MEDIUM
    action: str
    message: str
    status_code: int

    def get_message(self) -> str:
        return f"Failed to {self.action}. HTTP status code {self.status_code}: {self.message}"


@dataclass(frozen=True)
class IgnoredValueWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    name: str
    value: str
    reason: str

    def get_message(self) -> str:
        return f"Ignoring value {self.value!r} for {self.name!r}: {self.reason!r}"


@dataclass(frozen=True)
class LimitedAccessWarning(ToolkitWarning):
    severity: ClassVar[SeverityLevel] = SeverityLevel.LOW
    message: str

    def get_message(self) -> str:
        return self.message



import sys
from types import MappingProxyType
from typing import Any, ClassVar, Literal, cast

from pydantic import Field, ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.constants import (
    DM_EXTERNAL_ID_PATTERN,
    SPACE_FORMAT_PATTERN,
)
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import BaseModelResource

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class ContainerReference(BaseModelResource):
    type: Literal["container"] = "container"
    space: str = Field(
        description="Id of the space hosting (containing) the container.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the container.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )


class ConstraintDefinition(BaseModelResource):
    constraint_type: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_constraint_definition_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, ConstraintDefinition):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid constraint definition data '{type(data)}' expected dict")

        if cls is not ConstraintDefinition:
            data_copy = dict(data)
            data_copy.pop("constraintType", None)
            return handler(data_copy)

        constraint_type = data.get("constraintType")
        if constraint_type is None:
            raise ValueError("Missing 'constraintType' field in constraint data")
        if constraint_type not in _CONSTRAINT_DEFINITION_CLASS_BY_TYPE:
            raise ValueError(
                f"invalid destination type '{constraint_type}'. Expected one of {humanize_collection(_CONSTRAINT_DEFINITION_CLASS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _CONSTRAINT_DEFINITION_CLASS_BY_TYPE[constraint_type]
        data_copy = dict(data)
        data_copy.pop("constraintType")

        return cast(Self, cls_.model_validate(data_copy))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def serialize_constrain_definition(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["constraintType"] = self.__class__.constraint_type
        return serialized_data


class UniquenessConstraintDefinition(ConstraintDefinition):
    constraint_type = "uniqueness"
    properties: list[str] = Field(description="List of properties included in the constraint.")
    by_space: bool | None = Field(default=None, description="Whether to make the constraint space-specific.")


class RequiresConstraintDefinition(ConstraintDefinition):
    constraint_type = "requires"
    require: ContainerReference = Field(description="Reference to an existing container.")


_CONSTRAINT_DEFINITION_CLASS_BY_TYPE: MappingProxyType[str, type[ConstraintDefinition]] = MappingProxyType(
    {c.constraint_type: c for c in ConstraintDefinition.__subclasses__()}
)


class IndexDefinition(BaseModelResource):
    index_type: ClassVar[str]
    properties: list[str] = Field(description="List of properties to define the index across.")

    @model_validator(mode="wrap")
    @classmethod
    def find_index_definition_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, IndexDefinition):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid index definition data '{type(data)}' expected dict")

        if cls is not IndexDefinition:
            data_copy = dict(data)
            data_copy.pop("indexType", None)
            return handler(data_copy)

        index_type = data.get("indexType")
        if index_type is None:
            raise ValueError("Missing 'indexType' field in index definition data")
        if index_type not in _INDEX_DEFINITION_CLASS_BY_TYPE:
            raise ValueError(
                f"invalid index type '{index_type}'. Expected one of {humanize_collection(_INDEX_DEFINITION_CLASS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _INDEX_DEFINITION_CLASS_BY_TYPE[index_type]
        data_copy = dict(data)
        data_copy.pop("indexType")
        return cast(Self, cls_.model_validate(data_copy))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def serialize_index(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["indexType"] = self.__class__.index_type
        return serialized_data


class BtreeIndex(IndexDefinition):
    index_type = "btree"
    by_space: bool | None = Field(default=None, description="Whether to make the index space-specific.")
    cursorable: bool | None = Field(
        default=None, description="Whether the index can be used for cursor-based pagination."
    )


class InvertedIndex(IndexDefinition):
    index_type = "inverted"


_INDEX_DEFINITION_CLASS_BY_TYPE: MappingProxyType[str, type[IndexDefinition]] = MappingProxyType(
    {c.index_type: c for c in IndexDefinition.__subclasses__()}
)


class PropertyTypeDefinition(BaseModelResource):
    type: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_property_type_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, PropertyTypeDefinition):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid property type data '{type(data)}' expected dict")

        if cls is not PropertyTypeDefinition:
            data_copy = dict(data)
            data_copy.pop("type", None)
            return handler(data_copy)

        property_type = data.get("type")
        if property_type is None:
            raise ValueError("Missing 'type' field in property type data")
        if property_type not in _PROPERTY_TYPE_CLASS_BY_TYPE:
            raise ValueError(
                f"invalid property type '{property_type}'. Expected one of {humanize_collection(_PROPERTY_TYPE_CLASS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _PROPERTY_TYPE_CLASS_BY_TYPE[property_type]
        data_copy = dict(data)
        data_copy.pop("type")
        return cast(Self, cls_.model_validate(data_copy))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def serialize_property_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["type"] = self.__class__.type
        return serialized_data


class ListablePropertyTypeDefinition(PropertyTypeDefinition):
    type: ClassVar[str]
    list: bool | None = Field(
        default=None,
        description="Specifies that the data type is a list of values.",
    )
    max_list_size: int | None = Field(
        default=None,
        description="Specifies the maximum number of values in the list",
    )


class TextProperty(ListablePropertyTypeDefinition):
    type = "text"
    collation: str | None = Field(
        default=None,
        description="he set of language specific rules - used when sorting text fields.",
    )


class FloatPrimitiveProperty(ListablePropertyTypeDefinition):
    type: ClassVar[Literal["float32", "float64"]]
    unit: dict[Literal["externalId", "sourceUnit"], str] | None = Field(
        default=None,
        description="The unit of the data stored in this property.",
    )


class Float32PrimitiveProperty(FloatPrimitiveProperty):
    type = "float32"


class Float64PrimitiveProperty(FloatPrimitiveProperty):
    type = "float64"


class BooleanPrimitiveProperty(ListablePropertyTypeDefinition):
    type = "boolean"


class Int32PrimitiveProperty(ListablePropertyTypeDefinition):
    type = "int32"


class Int64PrimitiveProperty(ListablePropertyTypeDefinition):
    type = "int64"


class TimestampPrimitiveProperty(ListablePropertyTypeDefinition):
    type = "timestamp"


class DatePrimitiveProperty(ListablePropertyTypeDefinition):
    type = "date"


class JSONPrimitiveProperty(ListablePropertyTypeDefinition):
    type = "json"


class TimeseriesCDFExternalIdReference(ListablePropertyTypeDefinition):
    type = "timeseries"


class FileCDFExternalIdReference(ListablePropertyTypeDefinition):
    type = "file"


class SequenceCDFExternalIdReference(ListablePropertyTypeDefinition):
    type = "sequence"


class DirectNodeRelation(ListablePropertyTypeDefinition):
    type = "direct"
    container: ContainerReference | None = Field(
        default=None,
        description="The (optional) required type for the node the direct relation points to.",
    )


class EnumProperty(PropertyTypeDefinition):
    type = "enum"
    unknown_value: str | None = Field(
        default=None,
        description="The value to use when the enum value is unknown.",
    )
    values: dict[str, dict[Literal["name", "description"], str]] = Field(
        description="A set of all possible values for the enum property."
    )


def get_all_property_type_leaf_classes(base_class: type[PropertyTypeDefinition]) -> list:
    subclasses = base_class.__subclasses__()
    result = []

    if not subclasses:
        if base_class is not PropertyTypeDefinition:
            result.append(base_class)
    else:
        for subclass in subclasses:
            result.extend(get_all_property_type_leaf_classes(subclass))

    return result


_PROPERTY_TYPE_CLASS_BY_TYPE: MappingProxyType[str, type[PropertyTypeDefinition]] = MappingProxyType(
    {
        cls.type: cls
        for cls in get_all_property_type_leaf_classes(PropertyTypeDefinition)
        if hasattr(cls, "type") and cls.type is not None
    }
)


class ContainerPropertyDefinition(BaseModelResource):
    immutable: bool | None = Field(
        default=None,
        description="Should updates to this property be rejected after the initial population?",
    )
    nullable: bool | None = Field(
        default=None,
        description="Does this property need to be set to a value, or not?",
    )
    auto_increment: bool | None = Field(
        default=None,
        description="Increment the property based on its highest current value (max value).",
    )
    default_value: str | int | bool | dict | None = Field(
        default=None,
        description="Default value to use when you do not specify a value for the property.",
    )
    description: str | None = Field(
        default=None,
        description="Description of the content and suggested use for this property.",
        max_length=1024,
    )
    name: str | None = Field(
        default=None,
        description="Readable property name.",
        max_length=255,
    )
    type: PropertyTypeDefinition = Field(description="The type of data you can store in this property.")

    @model_serializer(mode="wrap")
    def serialize_type(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        serialized_data = handler(self)
        if self.type:
            serialized_data["type"] = self.type.model_dump(**vars(info))
        return serialized_data



from typing import cast

from cognite.client import CogniteClient
from rich.console import Console

from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient

from .api.canvas import CanvasAPI
from .api.charts import ChartsAPI
from .api.dml import DMLAPI
from .api.extended_data_modeling import ExtendedDataModelingAPI
from .api.extended_files import ExtendedFileMetadataAPI
from .api.extended_functions import ExtendedFunctionsAPI
from .api.extended_raw import ExtendedRawAPI
from .api.extended_timeseries import ExtendedTimeSeriesAPI
from .api.infield import InfieldAPI
from .api.lookup import LookUpGroup
from .api.migration import MigrationAPI
from .api.project import ProjectAPI
from .api.robotics import RoboticsAPI
from .api.search import SearchAPI
from .api.streams import StreamsAPI
from .api.token import TokenAPI
from .api.verify import VerifyAPI
from .config import ToolkitClientConfig


class ToolkitClient(CogniteClient):
    def __init__(self, config: ToolkitClientConfig | None = None, enable_set_pending_ids: bool = False) -> None:
        super().__init__(config=config)
        http_client = HTTPClient(self.config)
        toolkit_config = ToolkitClientConfig.from_client_config(self.config)
        self.console = Console()
        self.search = SearchAPI(self._config, self._API_VERSION, self)
        self.robotics = RoboticsAPI(self._config, self._API_VERSION, self)
        self.dml = DMLAPI(self._config, self._API_VERSION, self)
        self.verify = VerifyAPI(self._config, self._API_VERSION, self)
        self.lookup = LookUpGroup(self._config, self._API_VERSION, self, self.console)
        self.functions: ExtendedFunctionsAPI = ExtendedFunctionsAPI(toolkit_config, self._API_VERSION, self)
        self.data_modeling: ExtendedDataModelingAPI = ExtendedDataModelingAPI(self._config, self._API_VERSION, self)
        if enable_set_pending_ids:
            self.time_series: ExtendedTimeSeriesAPI = ExtendedTimeSeriesAPI(self._config, self._API_VERSION, self)
            self.files: ExtendedFileMetadataAPI = ExtendedFileMetadataAPI(self._config, self._API_VERSION, self)
        self.raw: ExtendedRawAPI = ExtendedRawAPI(self._config, self._API_VERSION, self)
        self.canvas = CanvasAPI(self.data_modeling.instances)
        self.migration = MigrationAPI(self.data_modeling.instances)
        self.token = TokenAPI(self)
        self.charts = ChartsAPI(self._config, self._API_VERSION, self)
        self.project = ProjectAPI(config=toolkit_config, cognite_client=self)
        self.infield = InfieldAPI(http_client, self.console)
        self.streams = StreamsAPI(http_client, self.console)

    @property
    def config(self) -> ToolkitClientConfig:
        """Returns a config object containing the configuration for the current client.

        Returns:
            ToolkitClientConfig: The configuration object.
        """
        return cast(ToolkitClientConfig, self._config)



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import difflib
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import capabilities as cap
from cognite.client.data_classes.capabilities import (
    Capability,
    GroupsAcl,
    SecurityCategoriesAcl,
)
from cognite.client.data_classes.iam import (
    Group,
    GroupList,
    GroupWrite,
    GroupWriteList,
    SecurityCategory,
    SecurityCategoryList,
    SecurityCategoryWrite,
    SecurityCategoryWriteList,
)
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console
from rich.markup import escape

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitWrongResourceError
from cognite_toolkit._cdf_tk.resource_classes import GroupYAML, SecurityCategoriesYAML
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
    MediumSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, hash_dict


@dataclass
class _ReplaceMethod:
    """This is a small helper class used in the
    lookup and replace in the ACL scoped ids"""

    lookup_method: Callable[[str, bool], int | None]
    reverse_lookup_method: Callable[[int], str | None]
    id_name: str


class GroupCRUD(ResourceCRUD[str, GroupWrite, Group, GroupWriteList, GroupList]):
    folder_name = "auth"
    filename_pattern = r"^(?!.*SecurityCategory$).*"
    kind = "Group"
    resource_cls = Group
    resource_write_cls = GroupWrite
    list_cls = GroupList
    list_write_cls = GroupWriteList
    yaml_cls = GroupYAML
    resource_scopes = frozenset(
        {
            cap.IDScope,
            cap.SpaceIDScope,
            cap.DataSetScope,
            cap.TableScope,
            cap.AssetRootIDScope,
            cap.ExtractionPipelineScope,
            cap.IDScopeLowerCase,
        }
    )
    resource_scope_names = frozenset({scope._scope_name for scope in resource_scopes})  # type: ignore[attr-defined]
    _doc_url = "Groups/operation/createGroups"

    def __init__(
        self,
        client: ToolkitClient,
        build_dir: Path | None,
        console: Console | None,
        target_scopes: Literal[
            "all_scoped_only",
            "resource_scoped_only",
        ] = "all_scoped_only",
    ):
        super().__init__(client, build_dir, console)
        self.target_scopes = target_scopes

    @property
    def display_name(self) -> str:
        return f"groups({self.target_scopes.removesuffix('_only')})"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[GroupWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
            ]
            if read_only
            else [
                GroupsAcl.Action.Read,
                GroupsAcl.Action.List,
                GroupsAcl.Action.Create,
                GroupsAcl.Action.Delete,
                GroupsAcl.Action.Update,
            ]
        )

        return GroupsAcl(
            actions,
            GroupsAcl.Scope.All(),
        )

    @classmethod
    def get_id(cls, item: GroupWrite | Group | dict) -> str:
        if isinstance(item, dict):
            return item["name"]
        return item.name

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"name": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        from .classic import AssetCRUD
        from .data_organization import DataSetsCRUD
        from .datamodel import SpaceCRUD
        from .extraction_pipeline import ExtractionPipelineCRUD
        from .location import LocationFilterCRUD
        from .raw import RawDatabaseCRUD, RawTableCRUD
        from .timeseries import TimeSeriesCRUD

        for capability in item.get("capabilities", []):
            for acl, content in capability.items():
                if scope := content.get("scope", {}):
                    if space_ids := scope.get(cap.SpaceIDScope._scope_name, []):
                        if isinstance(space_ids, dict) and "spaceIds" in space_ids:
                            for space_id in space_ids["spaceIds"]:
                                yield SpaceCRUD, space_id
                    if data_set_ids := scope.get(cap.DataSetScope._scope_name, []):
                        if isinstance(data_set_ids, dict) and "ids" in data_set_ids:
                            for data_set_id in data_set_ids["ids"]:
                                yield DataSetsCRUD, data_set_id
                    if table_ids := scope.get(cap.TableScope._scope_name, []):
                        for db_name, tables in table_ids.get("dbsToTables", {}).items():
                            yield RawDatabaseCRUD, RawDatabase(db_name)
                            if isinstance(tables, list):
                                yield from ((RawTableCRUD, RawTable(db_name, table)) for table in tables)
                            elif isinstance(tables, dict) and "tables" in tables:
                                for table in tables["tables"]:
                                    yield RawTableCRUD, RawTable(db_name, table)
                    if extraction_pipeline_ids := scope.get(cap.ExtractionPipelineScope._scope_name, []):
                        if isinstance(extraction_pipeline_ids, dict) and "ids" in extraction_pipeline_ids:
                            for extraction_pipeline_id in extraction_pipeline_ids["ids"]:
                                yield ExtractionPipelineCRUD, extraction_pipeline_id
                    if asset_root_ids := scope.get(cap.AssetRootIDScope._scope_name, []):
                        if isinstance(asset_root_ids, dict) and "rootIds" in asset_root_ids:
                            for asset_root_id in asset_root_ids["rootIds"]:
                                yield AssetCRUD, asset_root_id
                    if (ids := scope.get(cap.IDScope._scope_name, [])) or (
                        ids := scope.get(cap.IDScopeLowerCase._scope_name, [])
                    ):
                        loader: type[ResourceCRUD] | None = None
                        if acl == cap.DataSetsAcl._capability_name:
                            loader = DataSetsCRUD
                        elif acl == cap.ExtractionPipelinesAcl._capability_name:
                            loader = ExtractionPipelineCRUD
                        elif acl == cap.TimeSeriesAcl._capability_name:
                            loader = TimeSeriesCRUD
                        elif acl == cap.SecurityCategoriesAcl._capability_name:
                            loader = SecurityCategoryCRUD
                        elif acl == cap.LocationFiltersAcl._capability_name:
                            loader = LocationFilterCRUD
                        if loader is not None and isinstance(ids, dict) and "ids" in ids:
                            for id_ in ids["ids"]:
                                yield loader, id_

    def _substitute_scope_ids(self, group: dict[str, Any], is_dry_run: bool, reverse: bool = False) -> dict[str, Any]:
        replace_method_by_acl = self._create_replace_method_by_acl_and_scope()

        for capability in group.get("capabilities", []):
            for acl, values in capability.items():
                scope = values.get("scope", {})
                if len(scope) != 1:
                    # This will raise an error when the group is loaded.
                    continue
                scope_name, _scope_content = next(iter(scope.items()))

                if (acl, scope_name) in replace_method_by_acl:
                    replace_method = replace_method_by_acl[(acl, scope_name)]
                elif scope_name in replace_method_by_acl:
                    replace_method = replace_method_by_acl[scope_name]
                else:
                    continue
                if ids := scope.get(scope_name, {}).get(replace_method.id_name, []):
                    if reverse:
                        cdf_ids = (
                            replace_method.reverse_lookup_method(int_id) if isinstance(int_id, int) else int_id
                            for int_id in ids
                        )
                        values["scope"][scope_name][replace_method.id_name] = [
                            id_ for id_ in cdf_ids if id_ is not None
                        ]
                    else:
                        values["scope"][scope_name][replace_method.id_name] = [
                            replace_method.lookup_method(ext_id, is_dry_run) if isinstance(ext_id, str) else ext_id
                            for ext_id in ids
                        ]
        return group

    def _create_replace_method_by_acl_and_scope(self) -> dict[tuple[str, str] | str, _ReplaceMethod]:
        source = {
            (cap.DataSetsAcl, cap.DataSetsAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            (cap.ExtractionPipelinesAcl, cap.ExtractionPipelinesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            (cap.LocationFiltersAcl, cap.LocationFiltersAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.location_filters.id,
                self.client.lookup.location_filters.external_id,
                id_name="ids",
            ),
            (cap.SecurityCategoriesAcl, cap.SecurityCategoriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.security_categories.id,
                self.client.lookup.security_categories.external_id,
                id_name="ids",
            ),
            (cap.TimeSeriesAcl, cap.TimeSeriesAcl.Scope.ID): _ReplaceMethod(
                self.client.lookup.time_series.id,
                self.client.lookup.time_series.external_id,
                id_name="ids",
            ),
            cap.DataSetScope: _ReplaceMethod(
                self.client.lookup.data_sets.id,
                self.client.lookup.data_sets.external_id,
                id_name="ids",
            ),
            cap.ExtractionPipelineScope: _ReplaceMethod(
                self.client.lookup.extraction_pipelines.id,
                self.client.lookup.extraction_pipelines.external_id,
                id_name="ids",
            ),
            cap.AssetRootIDScope: _ReplaceMethod(
                self.client.lookup.assets.id,
                self.client.lookup.assets.external_id,
                id_name="rootIds",
            ),
        }
        # Trick to avoid writing _capability_name and _scope_name for each entry.
        return {
            (key[0]._capability_name, key[1]._scope_name) if isinstance(key, tuple) else key._scope_name: method  # type: ignore[attr-defined]
            for key, method in source.items()
        }

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> GroupWrite:
        is_resource_scoped = any(
            any(scope_name in capability.get(acl, {}).get("scope", {}) for scope_name in self.resource_scope_names)
            for capability in resource.get("capabilities", [])
            for acl in capability
        )

        if self.target_scopes == "all_scoped_only" and is_resource_scoped:
            raise ToolkitWrongResourceError()

        if self.target_scopes == "resource_scoped_only" and not is_resource_scoped:
            raise ToolkitWrongResourceError()

        substituted = self._substitute_scope_ids(resource, is_dry_run)
        try:
            loaded = GroupWrite.load(substituted)
        except ValueError:
            # The GroupWrite class in the SDK will raise a ValueError if the ACI or scope is not valid or unknown.
            loaded = GroupWrite._load(substituted, allow_unknown=True)
            for capability in loaded.capabilities or []:
                if isinstance(capability, cap.UnknownAcl):
                    msg = (
                        f"In group {loaded.name!r}, unknown capability found: {capability.capability_name!r}.\n"
                        "Will proceed with group creation and let the API validate the capability."
                    )
                    if matches := difflib.get_close_matches(capability.capability_name, cap.ALL_CAPABILITIES):
                        msg += f"\nIf the API rejects the capability, could it be that you meant on of: {matches}?"
                    prefix, warning_msg = MediumSeverityWarning(msg).print_prepare()
                    print(prefix, warning_msg)

        return loaded

    def dump_resource(self, resource: Group, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        if not dumped.get("sourceId") and "sourceId" not in local:
            dumped.pop("sourceId", None)
        # RAWAcls are not returned by the API following the spec.
        # If you have a table scoped RAW ACL the spec, and thus user will input
        # tableScope:
        #   dbsToTables
        #    db1:
        #    - tables1
        #    - tables2
        # While the API will return
        # tableScope:
        #   dbsToTables:
        #   db1:
        #     tables: [tables1, tables2]
        # Note the extra keyword 'tables' in the API response.
        for capability in dumped.get("capabilities", []):
            for acl, content in capability.items():
                if acl != cap.RawAcl._capability_name:
                    continue
                if scope := content.get("scope", {}):
                    if table_scope := scope.get(cap.TableScope._scope_name, {}):
                        db_to_tables = table_scope.get("dbsToTables", {})
                        if not db_to_tables:
                            continue
                        for db_name in list(db_to_tables.keys()):
                            tables = db_to_tables[db_name]
                            if isinstance(tables, dict) and "tables" in tables:
                                db_to_tables[db_name] = tables["tables"]
        # When you dump a CDF Group, all the referenced resources should be available in CDF.
        return self._substitute_scope_ids(dumped, is_dry_run=False, reverse=True)

    def create(self, items: Sequence[GroupWrite]) -> GroupList:
        if len(items) == 0:
            return GroupList([])
        return self._create_with_fallback(items, action="create")

    def update(self, items: Sequence[GroupWrite]) -> GroupList:
        # We MUST retrieve all the old groups BEFORE we add the new, if not the new will be deleted
        old_groups = self.client.iam.groups.list(all=True)
        created = self._create_with_fallback(items, action="update")
        created_names = {g.name for g in created}
        to_delete = GroupList([group for group in old_groups if group.name in created_names])
        if to_delete:
            self._delete(to_delete, check_own_principal=False)
        return created

    def _create_with_fallback(self, items: Sequence[GroupWrite], action: Literal["create", "update"]) -> GroupList:
        try:
            return self.client.iam.groups.create(items)
        except CogniteAPIError as e:
            if not (e.code == 400 and "buffer" in e.message.lower() and len(items) > 1):
                raise e
            # Fallback to create one by one
            created_list = GroupList([])
            for item in items:
                try:
                    created = self.client.iam.groups.create(item)
                except CogniteAPIError as e:
                    HighSeverityWarning(f"Failed to {action} group {item.name}. Error: {escape(str(e))}").print_warning(
                        include_timestamp=True, console=self.console
                    )
                else:
                    created_list.append(created)
            return created_list

    def retrieve(self, ids: SequenceNotStr[str]) -> GroupList:
        id_set = set(ids)
        remote = self.client.iam.groups.list(all=True)
        found = [g for g in remote if g.name in id_set]
        return GroupList(found)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        return self._delete(self.retrieve(ids), check_own_principal=True)

    def _delete(self, delete_candidates: GroupList, check_own_principal: bool = True) -> int:
        if check_own_principal:
            print_fun = self.console.print if self.console else print
            try:
                # Let's prevent that we delete groups we belong to
                my_groups = self.client.iam.groups.list()
            except CogniteAPIError as e:
                print_fun(
                    f"[bold red]ERROR:[/] Failed to retrieve the current service principal's groups. Aborting group deletion.\n{e}"
                )
                return 0
            my_source_ids = {g.source_id for g in my_groups if g.source_id}
        else:
            my_source_ids = set()

        to_delete: list[int] = []
        counts_by_name: dict[str, int] = defaultdict(int)
        for group in delete_candidates:
            if group.source_id in my_source_ids:
                HighSeverityWarning(
                    f"Not deleting group {group.name} with sourceId {group.source_id} as it is used by"
                    f"the current service principal. If you want to delete this group, you must do it manually."
                ).print_warning(console=self.console)
            else:
                to_delete.append(group.id)
                counts_by_name[group.name] += 1
        if duplicates := {name for name, count in counts_by_name.items() if count > 1}:
            MediumSeverityWarning(
                f"The following names are used by multiple groups (all will be deleted): {duplicates}"
            ).print_warning(console=self.console)

        failed_deletes = []
        error_str = ""
        try:
            self.client.iam.groups.delete(to_delete)
        except CogniteNotFoundError:
            # Fallback to delete one by one
            for delete_item_id in to_delete:
                try:
                    self.client.iam.groups.delete(delete_item_id)
                except CogniteNotFoundError:
                    # If the group is already deleted, we can ignore the error
                    ...
                except CogniteAPIError as e:
                    error_str = str(e)
                    failed_deletes.append(delete_item_id)
        except CogniteAPIError as e:
            error_str = str(e)
            failed_deletes.extend(to_delete)
        if failed_deletes:
            MediumSeverityWarning(
                f"Failed to delete groups: {humanize_collection(to_delete)}. "
                "These must be deleted manually in the Fusion UI."
                f"Error: {escape(error_str)}"
            ).print_warning(include_timestamp=True, console=self.console)
        return len(to_delete) - len(failed_deletes)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Group]:
        return self.client.iam.groups.list(all=True)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("capabilities",):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        elif json_path[0] == "capabilities":
            # All sublist inside capabilities are hashable
            return diff_list_hashable(local, cdf)
        elif json_path == ("members",):
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


@final
class GroupAllScopedCRUD(GroupCRUD):
    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console, "all_scoped_only")

    @property
    def display_name(self) -> str:
        return "all-scoped groups"


@final
class SecurityCategoryCRUD(
    ResourceCRUD[str, SecurityCategoryWrite, SecurityCategory, SecurityCategoryWriteList, SecurityCategoryList]
):
    filename_pattern = r"^.*SecurityCategory$"  # Matches all yaml files who's stem ends with *SecurityCategory.
    resource_cls = SecurityCategory
    resource_write_cls = SecurityCategoryWrite
    list_cls = SecurityCategoryList
    list_write_cls = SecurityCategoryWriteList
    kind = "SecurityCategory"
    yaml_cls = SecurityCategoriesYAML
    folder_name = "auth"
    dependencies = frozenset({GroupAllScopedCRUD})
    _doc_url = "Security-categories/operation/createSecurityCategories"

    @property
    def display_name(self) -> str:
        return "security categories"

    @classmethod
    def get_id(cls, item: SecurityCategoryWrite | SecurityCategory | dict) -> str:
        if isinstance(item, dict):
            return item["name"]
        return cast(str, item.name)

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"name": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SecurityCategoryWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.MemberOf,
            ]
            if read_only
            else [
                SecurityCategoriesAcl.Action.Create,
                SecurityCategoriesAcl.Action.Update,
                SecurityCategoriesAcl.Action.MemberOf,
                SecurityCategoriesAcl.Action.List,
                SecurityCategoriesAcl.Action.Delete,
            ]
        )

        return SecurityCategoriesAcl(
            actions,
            SecurityCategoriesAcl.Scope.All(),
        )

    def create(self, items: SecurityCategoryWriteList) -> SecurityCategoryList:
        return self.client.iam.security_categories.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SecurityCategoryList:
        names = set(ids)
        categories = self.client.iam.security_categories.list(limit=-1)
        return SecurityCategoryList([c for c in categories if c.name in names])

    def update(self, items: SecurityCategoryWriteList) -> SecurityCategoryList:
        items_by_name = {item.name: item for item in items}
        retrieved = self.retrieve(list(items_by_name.keys()))
        retrieved_by_name = {item.name: item for item in retrieved}
        new_items_by_name = {item.name: item for item in items if item.name not in retrieved_by_name}
        if new_items_by_name:
            created = self.client.iam.security_categories.create(list(new_items_by_name.values()))
            retrieved_by_name.update({item.name: item for item in created})
        return SecurityCategoryList([retrieved_by_name[name] for name in items_by_name])

    def delete(self, ids: SequenceNotStr[str]) -> int:
        retrieved = self.retrieve(ids)
        if retrieved:
            self.client.iam.security_categories.delete([item.id for item in retrieved if item.id])
        return len(retrieved)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[SecurityCategory]:
        return self.client.iam.security_categories.list(limit=-1)



import pytest
from cognite.client.data_classes.data_modeling import DataModelId, ViewId

from cognite_toolkit._cdf_tk.utils import (
    GraphQLParser,
)
from cognite_toolkit._cdf_tk.utils.graphql_parser import _Directive, _DirectiveTokens, _ViewDirective

SPACE = "sp_my_space"
DATA_MODEL = DataModelId(SPACE, "MyDataModel", "v1")
GraphQLTestCases = [
    pytest.param(
        '''"Log File"
type LogFile
  @view(
    rawFilter: {
      hasData: [
        {
          type: "container"
          space: "onse_logfile"
          externalId: "LogFile"
        }
      ]
    }  version: "1"
  ) {
  """
  Name of the CSV which delivers result/logs from ONSE
  @name CSV LogFile Name
  """
  csv_logfile_name: String
  """
  CDF address of the posted CSV log file from ONSE
  @name CSV LogFile Unique ID
  """
  csv_logfile_uniqueId: Float
  """
  General result from ONSE such as Success or Partial or Failure (Status)
  @name Status
  """
  status: String
  """
  Cryptic id from IX representing the Transmittal/transaction (key)
  @name Transaction id
  """
  transaction_id: String
  """
  Transmittal Number (Referemce/Identifier)
  @name Transmittal Number
  """
  transmittalNumber: String
  """
  Date and time when xml LogFile has been posted to CDF
  @name Uploaded to CDF  Time
  """
  uploaded_to_cdf: String
  """
  Name of the XML which delivers result/logs from ONSE
  @name XML LogFile Name
  """
  xml_logfile_name: String
  """
  CDF address of the posted XML log file from ONSE
  @name XML LogFile Unique ID
  """
  xml_logfile_uniqueId: Float
}

"Log File From ONSE"
type LogFile_DryRun
  @view(
    rawFilter: {
      hasData: [
        {
          type: "container"
          space: "onse_logfile"
          externalId: "LogFile_DryRun"
        }
      ]
    }  version: "1"
  ) {
  """
  Name of the CSV which delivers result/logs from ONSE
  @name CSV LogFile Name
  """
  csv_logfile_name: String
  """
  CDF address of the posted CSV log file from ONSE
  @name CSV LogFile Unique ID
  """
  csv_logfile_uniqueId: Float
  """
  General result from ONSE
  @name Status
  """
  status: String
  """
  Cryptic id from IX
  @name Transaction id
  """
  transaction_id: String
  """
  Transmittal Number (Referemce/Identifier)
  @name Transmittal Number
  """
  transmittalNumber: String
  """
  Date and time when xml LogFile has been posted to CDF
  @name Uploaded to CDF  Time
  """
  uploaded_to_cdf: String
  """
  Name of the XML which delivers result/logs from ONSE
  @name XML LogFile Name
  """
  xml_logfile_name: String
  """
  CDF address of the posted XML log file from ONSE
  @name XML LogFile Unique ID
  """
  xml_logfile_uniqueId: Float
}
''',
        DATA_MODEL,
        {ViewId(SPACE, "LogFile", "1"), ViewId(SPACE, "LogFile_DryRun", "1")},
        set(),
        id="Simple type with view and raw filter",
    ),
    pytest.param(
        """interface Creatable @view(space: "cdf_apps_shared", version: "v1") @import {
  visibility: String
  createdBy: CDF_User
  updatedBy: CDF_User
  isArchived: Boolean
}""",
        DATA_MODEL,
        set(),
        {ViewId("cdf_apps_shared", "Creatable", "v1")},
        id="Imported interface",
    ),
    pytest.param(
        """"@code WOOL"
type WorkOrderObjectListItem @import(dataModel: {externalId: "MaintenanceDOM", version: "2_2_0", space: "EDG-COR-ALL-DMD"}) {
  ...
}""",
        DATA_MODEL,
        set(),
        {DataModelId(space="EDG-COR-ALL-DMD", external_id="MaintenanceDOM", version="2_2_0")},
        id="Imported data model",
    ),
    pytest.param(
        '''
"""
@name        Notification
@description This is a SAP notification object.
@code        NOTI
"""
type Notification @container(indexes: [{identifier: "refCostCenterIndex", indexType: BTREE, fields: ["refCostCenter"]}, {identifier: "createdDateIndex", indexType: BTREE, fields: ["createdDate"], cursorable: true}]) @import(dataModel: {externalId: "MaintenanceDOM", version: "2_2_0", space: "EDG-COR-ALL-DMD"}) {
  ...
  }
''',
        DATA_MODEL,
        set(),
        {DataModelId(space="EDG-COR-ALL-DMD", external_id="MaintenanceDOM", version="2_2_0")},
        id="Imported data model with unused container directive",
    ),
    pytest.param(
        '''
    """
@code WKCC
@Description Work Center Category in SAP
"""
type WorkCenterCategory {
    """
    @name Name
    """
    name: String!
    """
    @name Description
    """
    description: String
    """
    @name Code
    """
    code: String
}''',
        DATA_MODEL,
        {ViewId(SPACE, "WorkCenterCategory", None)},
        set(),
        id="Simple type",
    ),
    pytest.param(
        """type Cdf3dConnectionProperties
  @import
  @view(space: "cdf_3d_schema", version: "1")
  @edge
  @container(
    constraints: [
      {
        identifier: "uniqueNodeRevisionConstraint"
        constraintType: UNIQUENESS
        fields: ["revisionId", "revisionNodeId"]
      }
    ]
  ) {
  revisionId: Int64!
  revisionNodeId: Int64!
}""",
        DATA_MODEL,
        set(),
        {ViewId("cdf_3d_schema", "Cdf3dConnectionProperties", "1")},
        id="Edge type",
    ),
    pytest.param(
        """type APM_User @view (version: "7") {
  name: String
  email: String
  lastSeen: Timestamp
  preferences: JSONObject
}""",
        DATA_MODEL,
        {ViewId(SPACE, "APM_User", "7")},
        set(),
        id="Simple type with version",
    ),
    pytest.param(
        """type UnitOfMeasurement
  @import(
    dataModel: {
      externalId: "CoreDOM"
      version: "1_0_18"
      space: "EDG-COR-ALL-DMD"
    }
  ) {
  name: String!
}""",
        DATA_MODEL,
        set(),
        {DataModelId(space="EDG-COR-ALL-DMD", external_id="CoreDOM", version="1_0_18")},
        id="No comma, only newline",
    ),
    pytest.param(
        '''"""
  @name Name }{ Breaks the parser
"""
type UnitOfMeasurement
  @import(
    dataModel: {
      externalId: "CoreDOM"
      version: "1_0_18"
      space: "EDG-COR-ALL-DMD"
    }
  ) {
  name: String!
}''',
        DATA_MODEL,
        set(),
        {DataModelId(space="EDG-COR-ALL-DMD", external_id="CoreDOM", version="1_0_18")},
        id="Ignore comments",
    ),
    pytest.param(
        """type APM_Config @view {
  name: String
  appDataSpaceId: String
  appDataSpaceVersion: String
  customerDataSpaceId: String
  customerDataSpaceVersion: String
  featureConfiguration: JSONObject
  fieldConfiguration: JSONObject
  rootLocationsConfiguration: JSONObject
}""",
        DATA_MODEL,
        {ViewId(SPACE, "APM_Config", None)},
        set(),
        id="No version",
    ),
    pytest.param(
        """"Navigational aid for traversing CogniteCADModel instances"
type CogniteCADModel implements CogniteDescribable & Cognite3DModel
  @view(
    space: "cdf_cdm"
    version: "v1"
    rawFilter: {
      and: [
        {
          hasData: [
            {
              type: "container"
              space: "cdf_cdm_3d"
              externalId: "Cognite3DModel"
            }
          ]
        }
        {
          equals: {
            property: ["cdf_cdm_3d", "Cognite3DModel", "type"]
            value: "CAD"
          }
        }
      ]
    }
  )
  @import {
  "Name of the instance"
  name: String
  "Description of the instance"
  description: String
  "Text based labels for generic use, limited to 1000"
  tags: [String]
  "Alternative names for the node"
  aliases: [String]
  "CAD, PointCloud or Image360"
  type: Cognite3DModel_type
  "Thumbnail of the 3D model"
  thumbnail: CogniteFile
  "List of revisions for this CAD model"
  revisions: [CogniteCADRevision]
    @reverseDirectRelation(throughProperty: "model3D")
}""",
        DATA_MODEL,
        set(),
        {ViewId("cdf_cdm", "CogniteCADModel", "v1")},
        id="Setting custom filter on view",
    ),
    pytest.param(
        '''"""
@name Tag (Beta)
@code CTG
@Description Beta version only. Should not be used unless aligned with Celanese Data Governance Owner. Tag is an object designed for performing functional requirements and serving as a specification for equipment.
"""
type TagBeta @view (version: "7#") {
  name: String
  description: String
  aliases: [String]
  isActive: Boolean
  tagTypes: [TagType]
  tagClass: CfihosTagClass
  functionalLocation: FunctionalLocation # --> To be deprecated. Use functionalLocations instead
  functionalLocations: [FunctionalLocation]
  equipment: Equipment # --> To be deprecated. Use equipments instead
  equipments: [Equipment]
  reportingUnit: ReportingUnit # --> To be deprecated. Use reportingUnits instead
  reportingUnits: [ReportingUnit]
}''',
        DATA_MODEL,
        {ViewId(SPACE, "TagBeta", "7#")},
        set(),
        id="Type with comments",
    ),
    pytest.param(
        """type Alarm @view (rawFilter:
{equals : {property: ["PSI-COR-ALL-DMD", "AlarmEventMessage", "journal"], value: "Alarm"}}
)
{
  journal:String @mapping(container: "AlarmEventMessage")
  dataOwner:AlarmEventMessageDataOwner @mapping(container: "AlarmEventMessage")
  aemTimeStamp: Timestamp @mapping(container: "AlarmEventMessage")
  pointTag:String @mapping(container: "AlarmEventMessage")
  pointTagDesc:String @mapping(container: "AlarmEventMessage")
  psiUnit:String @mapping(container: "AlarmEventMessage")
  code:String @mapping(container: "AlarmEventMessage")
  alarmType:String @mapping(container: "AlarmEventMessage")
  alarmState:String @mapping(container: "AlarmEventMessage")
  priority:String @mapping(container: "AlarmEventMessage")
  limit:Int @mapping(container: "AlarmEventMessage")
  value:String @mapping(container: "AlarmEventMessage")
  sourceAlarm:String @mapping(container: "AlarmEventMessage")
  alarmEnableStatus:String @mapping(container: "AlarmEventMessage")
}""",
        DATA_MODEL,
        {ViewId(SPACE, "Alarm", None)},
        set(),
        id="Type with raw filter",
    ),
]

DirectiveTestCases = [
    pytest.param(
        """view(
    rawFilter: {
      hasData: [
        {
          type: "container"
          space: "onse_logfile"
          externalId: "LogFile"
        }
      ]
    }
    version: "1"
  )""",
        _ViewDirective(version="1"),
        id="View directive with raw filter and version",
    ),
    pytest.param(
        """view(
    space: "cdf_cdm"
    version: "v1"
    rawFilter: {
      and: [
        {
          hasData: [
            {
              type: "container"
              space: "cdf_cdm_3d"
              externalId: "Cognite3DModel"
            }
          ]
        }
        {
          equals: {
            property: ["cdf_cdm_3d", "Cognite3DModel", "type"]
            value: "CAD"
          }
        }
      ]
    }
  )""",
        _ViewDirective(space="cdf_cdm", version="v1"),
    ),
    pytest.param(
        """view (rawFilter:
{equals : {property: ["PSI-COR-ALL-DMD", "AlarmEventMessage", "journal"], value: "Alarm"}}
)""",
        _ViewDirective(),
    ),
]


class TestGraphQLParser:
    @pytest.mark.parametrize("raw, data_model_id, expected_views, dependencies", GraphQLTestCases)
    def test_parse(
        self, raw: str, data_model_id: DataModelId, expected_views: set[ViewId], dependencies: set[ViewId | DataModelId]
    ) -> None:
        parser = GraphQLParser(raw, data_model_id)

        actual_views = parser.get_views(include_version=True)
        assert expected_views == actual_views
        actual_dependencies = parser.get_dependencies(include_version=True)
        assert dependencies == actual_dependencies

    @pytest.mark.parametrize("string, expected", DirectiveTestCases)
    def test_create_directive(self, string: str, expected: _Directive) -> None:
        tokens = GraphQLParser._token_pattern.findall(string)
        actual = _DirectiveTokens(tokens).create()
        assert expected == actual



import pytest
from cognite.client.data_classes import (
    Database,
    Table,
    Transformation,
    TransformationDestination,
    TransformationPreviewResult,
)
from rich.spinner import Spinner

from cognite_toolkit._cdf_tk.client.data_classes.raw import (
    RawProfileColumns,
    RawProfileResults,
    StringProfile,
    StringProfileColumn,
)
from cognite_toolkit._cdf_tk.commands import ProfileRawCommand
from cognite_toolkit._cdf_tk.constants import MAX_ROW_ITERATION_RUN_QUERY
from tests.test_unit.approval_client import ApprovalToolkitClient


@pytest.fixture()
def raw_profile_results_single_column() -> RawProfileResults:
    return RawProfileResults(
        row_count=ProfileRawCommand.max_profile_raw_count,
        columns=RawProfileColumns(
            {
                "externalId": StringProfileColumn(
                    count=500,
                    null_count=ProfileRawCommand.max_profile_raw_count - 500,
                    string=StringProfile(
                        length_range=(1, 10),
                        distinct_count=500,
                        length_histogram=((1, 2), (100, 200)),
                        value_counts={"test1": 100, "test2": 200, "test3": 200},
                        count=500,
                    ),
                )
            }
        ),
        is_complete=False,
    )


class TestProfileCommand:
    @pytest.mark.parametrize(
        "this_row, last_row, expected",
        [
            pytest.param(
                ["MyTable", "13", "2", "-", "-", "-"],
                ["MyTable", "12", "1", "-", "-", "-"],
                ["", "13", "2", "-", "-", "-"],
                id="No change in table name",
            ),
            pytest.param(
                ["MyTable", "13", "2", "-", "-", "-"],
                [],
                ["MyTable", "13", "2", "-", "-", "-"],
                id="Empty last row",
            ),
            pytest.param(
                ["MyTable", "13", "2", "my_transformation", "assets", "upsert"],
                ["OtherTable", "13", "2", "my_transformation", "assets", "upsert"],
                ["MyTable", "13", "2", "my_transformation", "assets", "upsert"],
                id="Change in table name with transformation",
            ),
            pytest.param(
                [
                    "MyTable",
                    Spinner(**ProfileRawCommand.spinner_args),
                    Spinner(**ProfileRawCommand.spinner_args),
                    "-",
                    "-",
                    "-",
                ],
                [
                    "MyTable",
                    Spinner(**ProfileRawCommand.spinner_args),
                    Spinner(**ProfileRawCommand.spinner_args),
                    "-",
                    "-",
                    "-",
                ],
                ["", "", "", "", "", ""],
                id="Spinner in both rows",
            ),
        ],
    )
    def test_create_draw_row(
        self, this_row: list[str | Spinner], last_row: list[str | Spinner], expected: list[str | Spinner]
    ) -> None:
        draw_row = ProfileRawCommand._create_draw_row(this_row, last_row)

        assert draw_row == expected

    def test_profile_raw_command_fallback_row_count(
        self, toolkit_client_approval: ApprovalToolkitClient, raw_profile_results_single_column: RawProfileResults
    ) -> None:
        """Test that when there is more than 10 000 rows in the table,
        the fallback is to use the /transformations/preview endpoint to get the row count."""
        cmd = ProfileRawCommand(silent=True)
        row_count = MAX_ROW_ITERATION_RUN_QUERY
        toolkit_client_approval.append(
            Transformation,
            Transformation(
                external_id="MyTransformation",
                name="My Transformation",
                conflict_mode="update",
                query="SELECT externalId, name FROM `database`.`table`",
                destination=TransformationDestination(type="events"),
            ),
        )
        toolkit_client_approval.append(Database, Database("database"))
        toolkit_client_approval.append(Table, Table("table"))

        toolkit_client_approval.mock_client.raw.profile.return_value = raw_profile_results_single_column
        toolkit_client_approval.mock_client.transformations.preview.return_value = TransformationPreviewResult(
            results=[{"row_count": row_count}]
        )

        results = cmd.raw(toolkit_client_approval.client, "events")

        assert len(results) == 1
        row = results[0]
        assert row[cmd.Columns.Rows] == f"≥{row_count:,}"
        assert row[cmd.Columns.Columns] == f"≥{raw_profile_results_single_column.column_count:,}"
        assert toolkit_client_approval.mock_client.transformations.preview.call_count == 1



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.robotics.frame import RobotFrameYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def valid_robot_frame_cases() -> Iterable:
    yield pytest.param(
        {
            "name": "Robot Base Frame",
            "externalId": "robot_base_frame",
            "transform": {
                "parentFrameExternalId": "world_frame",
                "translation": {"x": 10.5, "y": 20.3, "z": 0.0},
                "orientation": {"x": 0.0, "y": 0.0, "z": 0.707, "w": 0.707},
            },
        },
        id="full-frame-with-transform",
    )
    yield pytest.param(
        {
            "name": "World Frame",
            "externalId": "world_frame",
        },
        id="minimal-robot-frame",
    )


def invalid_robot_frame_cases() -> Iterable:
    yield pytest.param(
        {"name": "Missing External ID"},
        {"Missing required field: 'externalId'"},
        id="missing-required-field",
    )
    yield pytest.param(
        {
            "name": "Test Frame",
            "externalId": "frame_001",
            "unknownField": "value",
        },
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )


class TestRobotFrameYAML:
    @pytest.mark.parametrize("data", list(find_resources("Frame", "robotics")))
    def test_load_valid_robot_frame_file(self, data: dict[str, object]) -> None:
        loaded = RobotFrameYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", valid_robot_frame_cases())
    def test_load_valid_robot_frame(self, data: dict[str, object]) -> None:
        loaded = RobotFrameYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", invalid_robot_frame_cases())
    def test_invalid_robot_frame_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, RobotFrameYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors
